import os
import random

import pandas as pd

# Chemin vers le dossier contenant les images
dossier_images = r"D:\Alie\Documents\CollectesTwitter\AhashExps\small"

# Liste des noms de fichiers d'images
noms_images = os.listdir(dossier_images)

# Initialisation des listes pour les colonnes du dataframe
keys = []
descriptions = []
likes = []

# Parcourir chaque nom de fichier
for nom_image in noms_images:
    # Générer un nombre aléatoire entre 1 et 10 pour le nombre de fois où le nom sera ajouté dans la colonne "key"
    nb_ajouts = random.randint(1, 10)
    
    for _ in range(nb_ajouts):
        # Ajouter le nom du fichier dans la liste "keys"
        keys.append(nom_image)
        
        # Générer un texte français aléatoire avec un entier variable
        texte_aleatoire = "Texte français aléatoire {}".format(random.randint(1, 100))
        
        # Ajouter le texte aléatoire dans la liste "descriptions"
        descriptions.append(texte_aleatoire)
        
        # Générer un nombre aléatoire entre 1 et 1500 pour la colonne "likes"
        nb_likes = random.randint(1, 1500)
        
        # Ajouter le nombre aléatoire dans la liste "likes"
        likes.append(nb_likes)

# Créer le dataframe
df = pd.DataFrame({"key": keys, "description[string]": descriptions, "likes[number]": likes})

# Sauvegarder le dataframe dans un fichier CSV
df.to_csv("prop_file.csv", sep=";", index=False)



import asyncio
import json

import aiosqlite

from panoptic.core.project.project import Project
from panoptic.utils import clean_value

source_db_file = '/Users/david/Downloads/panoptic-edouard.db'
target_folder = '/Users/david/panoptic-projects/panoptic-edouard'


async def run():
    async with aiosqlite.connect(source_db_file) as db:
        # Execute a SELECT query
        computed_values = []
        instances = []
        folders = []
        properties = []
        instance_values = []
        image_values = []
        tags = []
        async with db.execute('SELECT * FROM computed_values') as cursor:
            async for row in cursor:
                row = list(row)
                computed_values.append(row)
        async with db.execute('SELECT * FROM images') as cursor:
            async for row in cursor:
                row = list(row)
                row[5] = row[5][8:]
                instances.append(row)
        async with db.execute('SELECT * FROM folders') as cursor:
            async for row in cursor:
                row = list(row)
                folders.append(row)
        async with db.execute('SELECT * FROM properties') as cursor:
            async for row in cursor:
                row = list(row)
                properties.append(row)
        async with db.execute('SELECT * FROM property_values') as cursor:
            async for row in cursor:
                row = list(row)
                if row[1] == -1:
                    row = [r for i, r in enumerate(row) if i != 1]
                    image_values.append(row)
                else:
                    print(row[2])
                    row = [r for i, r in enumerate(row) if i != 2]
                    instance_values.append(row)
        async with db.execute('SELECT * FROM tags') as cursor:
            async for row in cursor:
                row = list(row)
                row[4] = 0
                tags.append(row)

        project = Project(target_folder, [])
        await project.start()

        sha1_to_ahash = {v[0]: v[1] for v in computed_values}

        for folder in folders:
            await project.db.get_raw_db().import_folder(*folder)
        for instance in instances:
            ahash = sha1_to_ahash[instance[4]] if instance[4] in sha1_to_ahash else 'none'
            await project.db.get_raw_db().import_instance(*instance, ahash)
        for prop in properties:
            await project.db.get_raw_db().import_property(*prop)
        prop_index = {p.id: p for p in await project.db.get_properties()}

        for tag in tags:
            if tag[1] in prop_index:
                await project.db.get_raw_db().import_tag(*tag)

        for val in image_values:
            if val[0] not in prop_index:
                continue
            v = clean_value(prop_index[val[0]], val[2])
            if type(v) is not int:
                v = json.loads(v)
            if v:
                val[2] = v
                val[1] = [val[1]]
                await project.db.get_raw_db().set_image_property_value(*val)
        for val in instance_values:
            v = clean_value(prop_index[val[0]], val[2])
            if type(v) is not int:
                v = json.loads(v)
            if v:
                val[2] = v
                val[1] = [val[1]]
                await project.db.get_raw_db().set_instance_property_value(*val)
        await project.close()


asyncio.run(run())



software_db_version = 1

DB_VERSION = 'db_version'


def create_folders_table():
    query = """
    CREATE TABLE folders(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        path TEXT UNIQUE,
        name TEXT,
        parent INTEGER,
        FOREIGN KEY (parent) REFERENCES folders (id) ON DELETE CASCADE
    );
    """
    return query


def create_tabs_table():
    query = """
    CREATE TABLE tabs(
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        name TEXT,
        data JSON
    );
    """
    return query


def create_properties_table():
    query = """
    CREATE TABLE properties (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        name TEXT,
        type TEXT,
        mode TEXT DEFAULT 'id'
    );
    """
    return query




def create_instances_table():
    query = """
    CREATE TABLE instances (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        folder_id INTEGER NOT NULL,
        name TEXT NOT NULL,
        extension TEXT NOT NULL,
        sha1 TEXT NOT NULL,
        url TEXT NOT NULL,
        height INTEGER NOT NULL,
        width INTEGER NOT NULL,
        ahash TEXT NOT NULL
    );
    CREATE INDEX idx_image_filepath ON instances (folder_id, name, extension);
    CREATE INDEX idx_image_sha1 ON instances (sha1);
    """
    return query


def create_property_values_table():
    query = """
    CREATE TABLE property_values (
        property_id INTEGER NOT NULL,
        instance_id INTEGER NOT NULL,
        sha1 TEXT INTEGER NOT NULL,
        value JSON,
        PRIMARY KEY (property_id, instance_id, sha1),
        FOREIGN KEY (property_id) REFERENCES properties (id) ON DELETE CASCADE
    );
    """
    return query


def create_instance_property_values_table():
    query = """
    CREATE TABLE instance_property_values (
        property_id INTEGER NOT NULL,
        instance_id INTEGER NOT NULL,
        value JSON,
        
        PRIMARY KEY (property_id, instance_id),
        FOREIGN KEY (property_id) REFERENCES properties (id) ON DELETE CASCADE
    );
    """
    return query


def create_image_property_values_table():
    query = """
    CREATE TABLE image_property_values (
        property_id INTEGER NOT NULL,
        sha1 TEXT INTEGER NOT NULL,
        value JSON,
        PRIMARY KEY (property_id, sha1),
        FOREIGN KEY (property_id) REFERENCES properties (id) ON DELETE CASCADE
    );
    """
    return query


def create_tags_table():
    query = """
    CREATE TABLE tags (
        id INTEGER PRIMARY KEY AUTOINCREMENT,
        property_id INTEGER NOT NULL,
        value TEXT NOT NULL,
        parents JSON,
        color TEXT,
        FOREIGN KEY (property_id) REFERENCES properties (id) ON DELETE CASCADE
    );
    """
    return query


def create_panoptic_table():
    query = f"""
    CREATE TABLE panoptic (
        key TEXT PRIMARY KEY,
        value TEXT NOT NULL
    );
    INSERT INTO panoptic (key, value)
    VALUES ('{DB_VERSION}', '{software_db_version}');
    """
    return query


def create_vectors_table():
    query = f"""
    CREATE TABLE vectors (
        source TEXT,
        type TEXT,
        sha1 TEXT,
        data ARRAY,
        
        PRIMARY KEY (source, type, sha1)
    );
    """
    return query


def create_plugin_defaults_table():
    query = f"""
    CREATE TABLE plugin_defaults (
        name TEXT,
        base JSON,
        functions JSON,
        
        PRIMARY KEY (name)
    );
    """
    return query


def create_actions_table():
    query = """
    CREATE TABLE action_params (
        name TEXT PRIMARY KEY,
        value TEXT NOT NULL
    );
    """
    return query


# tables2 = {
#     'folders': create_folders_table(),
#     'tabs': create_tabs_table(),
#     'properties': create_properties_table(),
#     'images': create_images_table(),
#     'property_values': create_property_values_table(),
#     'tags': create_tags_table(),
#     'panoptic': create_panoptic_table(),
#     'vectors': create_vectors_table(),
#     'plugin_defaults': create_plugin_defaults_table(),
#     'action_params': create_actions_table()
# }

tables = {
    'panoptic': create_panoptic_table(),
    'folders': create_folders_table(),
    'instances': create_instances_table(),
    'properties': create_properties_table(),
    'image_property_values': create_image_property_values_table(),
    'instance_property_values': create_instance_property_values_table(),
    'tags': create_tags_table(),
    'vectors': create_vectors_table(),
    'plugin_defaults': create_plugin_defaults_table(),
    'action_params': create_actions_table(),
    'tabs': create_tabs_table(),
}



from panoptic.models import Property


class ComputedId:
    id = -1
    sha1 = -2
    ahash = -3
    folder = -4
    width = -5
    height = -6
    path = -7


id_property = Property(id=ComputedId.id, name='id', type='id', mode='id', computed=True)
sha1_property = Property(id=ComputedId.sha1, name='sha1', type='sha1', mode='sha1', computed=True)
ahash_property = Property(id=ComputedId.ahash, name='average hash', type='ahash', mode='sha1', computed=True)
folder_property = Property(id=ComputedId.folder, name='folder', type='folder', mode='id', computed=True)
width_property = Property(id=ComputedId.width, name='width', type='width', mode='sha1', computed=True)
height_property = Property(id=ComputedId.height, name='height', type='height', mode='sha1', computed=True)
path_property = Property(id=ComputedId.path, name='path', type='path', mode='id', computed=True)

computed_properties = {p.id: p for p in
                       [id_property, sha1_property, ahash_property, folder_property, width_property, height_property,
                        path_property]}



from typing import Callable, Awaitable

from panoptic.models import Instance
from panoptic.utils import EventListener


class ImportInstanceEvent(EventListener):
    def register(self, callback: Callable[[Instance], Awaitable[None]]):
        super().register(callback)

    def emit(self, event: Instance):
        super().emit(event)


class ProjectEvents:
    def __init__(self):
        self.import_instance = ImportInstanceEvent()



from __future__ import annotations

from typing import Any

from fastapi_camelcase import CamelModel

from .models import PropertyType, ActionParam, ActionContext, SetMode, ColumnOption


class ImagePayload(CamelModel):
    file_path: str


class PropertyPayload(CamelModel):
    name: str
    type: PropertyType
    mode: str = 'id'


class UpdatePropertyPayload(CamelModel):
    id: int
    name: str
    # id: PropertyType | None = None


class ExportPropertiesPayload(CamelModel):
    name: str | None = None
    properties: list[int] | None = None
    images: list[int] | None = None
    export_images: bool | None = None


class SetPropertyValuePayload(CamelModel):
    property_id: int
    instance_ids: list[int] | None = None
    value: Any


class SetTagPropertyValuePayload(SetPropertyValuePayload):
    mode: SetMode


class AddTagPayload(CamelModel):
    property_id: int
    value: str
    parent_id: int = None
    color: int = 0


class AddTagParentPayload(CamelModel):
    tag_id: int
    parent_id: int


class UpdateTagPayload(CamelModel):
    id: int
    value: str
    parent_id: list[int] | None = None
    color: int | None = None


class MakeClusterPayload(CamelModel):
    nb_groups: int = 50
    image_list: list[str] = []


class GetSimilarImagesPayload(CamelModel):
    sha1_list: list[str]


class GetSimilarImagesFromTextPayload(CamelModel):
    input_text: str


class ChangeProjectPayload(CamelModel):
    project: str


class StrPayload(CamelModel):
    value: str


class UpdateActionsPayload(CamelModel):
    updates: list[ActionParam]


class ExecuteActionPayload(CamelModel):
    action: str
    function: str = None
    context: ActionContext


class OptionsPayload(CamelModel):
    options: dict[int, ColumnOption]



import logging
import os
from sys import platform
from typing import Optional, Dict, Any

from fastapi import APIRouter, File, UploadFile
from fastapi.responses import ORJSONResponse
from pydantic import BaseModel
from starlette.responses import FileResponse

from panoptic.core.project.project import Project
from panoptic.models import Property, Tag, PropertyPayload, \
    SetPropertyValuePayload, AddTagPayload, \
    Tab, AddTagParentPayload, PropertyUpdate, \
    TagUpdate, Clusters, ActionContext, PluginDefaultParams, UpdateActionsPayload, VectorDescription, \
    ExecuteActionPayload, SetTagPropertyValuePayload, ImportOptions, OptionsPayload, ExportPropertiesPayload

project_router = APIRouter()

project: Project | None = None


def set_project(p: Project | None):
    global project
    project = p


# Route pour créer une property et l'insérer dans la table des properties
@project_router.post("/property")
async def create_property_route(payload: PropertyPayload) -> Property:
    return await project.db.add_property(payload.name, payload.type, payload.mode)


@project_router.get("/property")
async def get_properties_route() -> list[Property]:
    return await project.db.get_properties()


@project_router.patch("/property")
async def update_property_route(payload: PropertyUpdate) -> Property:
    return await project.db.update_property(payload)


@project_router.post('/property/file')
async def properties_by_file(file):
    pass


@project_router.post('/upload_file')
async def upload_file_route(file: UploadFile):
    await project.importer.upload_csv(file)
    return await project.importer.analyse_file()


@project_router.post('/import_file')
async def upload_file_route(req: OptionsPayload):
    return await project.importer.import_file(req.options)

#
# @project_router.post('/analyse_file')
# async def analyse_file_route():
#     return await project.importer.analyse_file(test.file)
#     # data = pd.read_csv(file.file, sep=";")
#     # return await read_properties_file(data)



@project_router.post('/export')
async def export_properties_route(req: ExportPropertiesPayload):
    await project.export_data(req.name, req.images, req.properties, req.export_images)
    return True



@project_router.delete('/property/{property_id}')
async def delete_property_route(property_id: str):
    await project.db.delete_property(property_id)
    return await project.db.get_properties()


@project_router.get("/images", response_class=ORJSONResponse)
async def get_all_images_route():
    images = await project.db.get_instances_with_properties()
    return ORJSONResponse(images)


@project_router.get('/images/{file_path:path}')
async def get_image(file_path: str):
    if platform == "linux" or platform == "linux2" or platform == "darwin":
        if not file_path.startswith('/'):
            file_path = '/' + file_path
    return FileResponse(path=file_path)


# Route pour ajouter une property à une image dans la table de jointure entre image et property
# On retourne le payload pour pouvoir valider l'update côté front
@project_router.post("/image_property")
async def set_property_values_route(payload: SetPropertyValuePayload):
    values = await project.db.set_property_values(property_id=payload.property_id,
                                                  instance_ids=payload.instance_ids,
                                                  value=payload.value)
    return values


@project_router.post('/set_tag_property_value')
async def set_tag_property_value(payload: SetTagPropertyValuePayload):
    print(payload)
    values = await project.db.set_tag_property_value(property_id=payload.property_id, instance_ids=payload.instance_ids,
                                                     value=payload.value, mode=payload.mode)
    return values


@project_router.post("/tags")
async def add_tag(payload: AddTagPayload) -> Tag:
    if not payload.parent_id:
        payload.parent_id = 0
    return await project.db.add_tag(payload.property_id, payload.value, payload.parent_id, payload.color)


@project_router.post("/tag/parent")
async def add_tag_parent(payload: AddTagParentPayload) -> Tag:
    return await project.db.add_tag_parent(payload.tag_id, payload.parent_id)


@project_router.get("/tags", response_class=ORJSONResponse)
async def get_tags_route(prop: Optional[int] = None):
    tags = await project.db.get_tags(prop)
    return ORJSONResponse(tags)


@project_router.patch("/tags")
async def update_tag_route(payload: TagUpdate) -> Tag:
    return await project.db.update_tag(payload)


@project_router.delete("/tags")
async def delete_tag_route(tag_id: int):
    return await project.db.delete_tag(tag_id)


@project_router.delete("/tags/parent")
async def delete_tag_parent_route(tag_id: int, parent_id: int):
    res = await project.db.delete_tag_parent(tag_id, parent_id)
    return res


@project_router.get("/folders")
async def get_folders_route():
    res = await project.db.get_folders()
    return res


# TODO
@project_router.get('/import_status')
async def get_import_status_route():
    if not project:
        return None
    res = project.get_status_update()
    return res


class PathRequest(BaseModel):
    path: str


class IdRequest(BaseModel):
    id: int


@project_router.post("/folders")
async def add_folder_route(path: PathRequest):
    # TODO: safe guards do avoid adding folder inside already imported folder. Also inverse direction
    nb_images = await project.import_folder(path.path)
    return await get_folders_route()


@project_router.post("/reimport_folder")
async def reimport_folder_route(req: IdRequest):
    folder = await project.db.get_folder(req.id)
    if not folder:
        raise Exception(f'Folder id does not exist [{req.id}]')
    await project.import_folder(folder.path)


@project_router.get("/tabs")
async def get_tabs_route():
    return await project.ui.get_tabs()


@project_router.post("/tab")
async def add_tab_route(tab: Tab):
    return await project.ui.add_tab(tab.data)


@project_router.patch("/tab")
async def update_tab_route(tab: Tab):
    return await project.ui.update_tab(tab)


@project_router.delete("/tab")
async def delete_tab_route(tab_id: int):
    return await project.ui.delete_tab(tab_id)


@project_router.get('/actions_description')
async def get_actions_description_route():
    res = project.action.get_actions_description()
    return res


@project_router.post('/actions_functions')
async def set_actions_update_route(actions_update: UpdateActionsPayload):
    await project.set_action_updates(actions_update.updates)
    return await get_actions_description_route()


@project_router.post('/action_execute')
async def execute_action_route(req: ExecuteActionPayload):
    res = await project.action.actions[req.action].call(req.context, function=req.function)
    return res


@project_router.get('/plugins_info')
async def get_plugins():
    res = await project.plugins_info()
    return res


@project_router.post('/plugins_params')
async def post_plugin_params_route(params: PluginDefaultParams):
    res = await project.set_plugin_default_params(params)
    return res


@project_router.get('/vectors_info')
async def get_vectors_info():
    vectors_description = await project.db.get_vectors_info()
    return vectors_description


@project_router.post('/default_vectors')
async def set_default_vectors(vector_description: VectorDescription):
    await project.db.set_default_vectors(vector_description)
    return await get_vectors_info()


#
# @project_router.post("/similar/text")
# async def get_similar_images_from_text_route(payload: GetSimilarImagesFromTextPayload) -> list:
#     return await get_similar_images_from_text(payload.input_text)


@project_router.get('/small/images/{file_path:path}')
async def get_image(file_path: str):
    path = os.path.join(project.base_path, 'mini', file_path)
    return FileResponse(path=path)


class EndpointFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        state = record.getMessage().find("/import_status") > -1
        small_img = record.getMessage().find("/small/images/") > -1
        img = record.getMessage().find("/images/") > -1
        return not (state or small_img or img)


# Filter out /endpoint
logging.getLogger("uvicorn.access").addFilter(EndpointFilter())



from __future__ import annotations

from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from panoptic.core.project.project import Project
from panoptic.core.task.task import Task

import importlib.util
import sys


def import_module_from_path(module_name, module_path):
    # Append the folder path to sys.path
    sys.path.append(module_path)

    # Construct the spec for the module
    spec = importlib.util.spec_from_file_location(module_name, module_path)

    # Import the module
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    # Finalize the loading
    spec.loader.exec_module(module)

    # Remove the path from sys.path to keep things clean
    # sys.path.remove(module_path)

    return module


class LoadPluginTask(Task):
    def __init__(self, project: Project, plugin_path: str):
        super().__init__()
        self.name = 'Load Plugin'
        self.project = project
        self.path = plugin_path

    async def run(self):
        path = Path(self.path)
        name = str(path.name)
        file_path = path / '__init__.py'
        plugin_module = await self._async(import_module_from_path, name, file_path)
        plugin = plugin_module.plugin_class(self.project, self.path)
        await plugin.start()
        self.project.plugins.append(plugin)

    async def run_if_last(self):
        await self.project.update_actions_from_db()
        self.project.plugin_loaded = True
        self.project.ui.update_counter.action += 1



# import atexit
# import json
# import logging
# import os
# import random
# import sys
# from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
# from typing import List, Any
#
#
# import pandas
# from fastapi import HTTPException
# from tqdm import tqdm
#
# from panoptic import compute
# from .db import db
# from panoptic.models import PropertyType, JSON, Tag, Property, Tags, Properties, \
#     UpdateTagPayload, UpdatePropertyPayload, Instance, PropertyValue, Clusters
# from .image_importer import ImageImporter
#
# nb_workers = 4
# executor = ThreadPoolExecutor(max_workers=nb_workers) if any(
#     [os.getenv('IS_DOCKER', False), sys.platform.startswith('linux')]) else ProcessPoolExecutor(
#     max_workers=nb_workers)
# atexit.register(executor.shutdown)
# importer = ImageImporter(executor)
#
#
# async def clear_import():
#     await importer.clear()
#
#
# async def create_property(name: str, property_type: PropertyType, mode='id') -> Property:
#     return await db.add_property(name, property_type.value, mode)
#
#
# async def update_property(payload: UpdatePropertyPayload) -> Property:
#     existing_property = await db.get_property_by_id(payload.id)
#     if not existing_property:
#         raise HTTPException(status_code=400, detail="Trying to modify non existent property")
#     new_property = existing_property.copy(update=payload.dict(exclude_unset=True))
#     await db.update_property(new_property)
#     return new_property
#
#
# async def get_properties() -> Properties:
#     properties = await db.get_properties()
#     return {prop.id: prop for prop in properties}
#
#
# async def delete_property(property_id: str):
#     await db.delete_property(property_id)
#
#
# async def set_property_values(property_id: int, value: Any, image_ids: List[int] = None, sha1s: List[int] = None):
#     if image_ids and sha1s:
#         raise TypeError('Only image_ids or sha1s should be given as keys. Never both')
#
#     prop = await db.get_property_by_id(property_id)
#     if prop.mode == 'id' and not image_ids:
#         raise TypeError(f'Property {property_id}: {prop.name} needs image ids as key [mode: {prop.mode}]')
#     if prop.mode == 'sha1' and not sha1s:
#         raise TypeError(f'Property {property_id}: {prop.name} needs sha1s as key [mode: {prop.mode}]')
#
#     if prop.id == PropertyType.tag or prop.id == PropertyType.multi_tags:
#         if value and not isinstance(value, list):
#             value = [int(value)]
#
#     if prop.id == PropertyType.checkbox:
#         value = True if value == 'true' or value is True else False
#
#     return await db.set_property_values(property_id, value, image_ids, sha1s)
#
#
# async def add_property_values(property_id: int, value: Any, image_ids: List[int] = None, sha1s: List[int] = None):
#     if image_ids and sha1s:
#         raise TypeError('Only image_ids or sha1s should be given as keys. Never both')
#
#     prop = await db.get_property_by_id(property_id)
#     if prop.mode == 'id' and not image_ids:
#         raise TypeError(f'Property {property_id}: {prop.name} needs image ids as key [mode: {prop.mode}]')
#     if prop.mode == 'sha1' and not sha1s:
#         raise TypeError(f'Property {property_id}: {prop.name} needs sha1s as key [mode: {prop.mode}]')
#
#     if prop.id != PropertyType.multi_tags:
#         raise TypeError('add_property_values is only supported for multi_tag properties')
#
#     if value and not isinstance(value, list):
#         value = [int(value)]
#
#     current_values = await db.get_property_values(property_ids=[property_id], image_ids=image_ids, sha1s=sha1s)
#     if image_ids:
#         value_index = {v.image_id: v.value for v in current_values}
#         [value_index.update({id_: []}) for id_ in image_ids if id_ not in value_index]
#         [value_index[id_].extend(value) for id_ in image_ids]
#         values = [list(set(value_index[id_])) for id_ in image_ids]
#         await db.set_multiple_property_values(property_id, values, image_ids)
#     else:
#         value_index = {v.sha1: v.value for v in current_values}
#         [value_index.update({sha1: []}) for sha1 in sha1s if sha1 not in value_index]
#         [value_index[sha1].extend(value) for sha1 in sha1s]
#         values = [list(set(value_index[sha1])) for sha1 in sha1s]
#         await db.set_multiple_property_values(property_id, values, sha1s)
#
#     updated_ids = image_ids
#     if not image_ids:
#         images = await db.get_images(sha1s=sha1s)
#         updated_ids = [img.id for img in images]
#     return updated_ids, value
#
#
# async def get_full_images(image_ids: List[int] = None) -> List[Instance]:
#     images = await db.get_images(image_ids)
#     sha1s = list({img.sha1 for img in images})
#     # get ids bound property values
#     property_values = await db.get_property_values(image_ids=image_ids)
#     # get sha1 bound property values when image_ids is set and therefore property_values only return id bound properties
#     if image_ids:
#         sha1s_values = await db.get_property_values(sha1s=sha1s)
#         property_values += sha1s_values
#     ahashs = await db.get_sha1_ahashs(sha1s=sha1s)
#     image_index = {img.id: img for img in images}
#
#     def assign_value(prop_value):
#         image_index[prop_value.image_id].properties[prop_value.property_id] = prop_value
#
#     # fill the index with ids bound property values
#     [assign_value(prop_value) for prop_value in property_values if prop_value.image_id >= 0]
#
#     sha1_properties = {}
#
#     def register_sha1_value(prop_value: PropertyValue):
#         if prop_value.sha1 not in sha1_properties:
#             sha1_properties[prop_value.sha1] = []
#         sha1_properties[prop_value.sha1].append(prop_value)
#
#     # populate the index of sha1 properties with prop_values that are bound to sha1s
#     [register_sha1_value(prop_values) for prop_values in property_values if prop_values.image_id < 0]
#
#     def assign_sha1_value(image_id, prop_value: PropertyValue):
#         image_index[image_id].properties[prop_value.property_id] = prop_value
#
#     # assign the sha1 bound property values to the images with corresponding sha1s
#     [assign_sha1_value(img.id, prop_value) for img in images if img.sha1 in sha1_properties for prop_value in
#      sha1_properties[img.sha1]]
#
#     [setattr(img, 'ahash', ahashs[img.sha1]) for img in images if img.sha1 in ahashs]
#     return images
#
#
# async def make_clusters(sensibility: float, sha1s: [str]) -> Clusters:
#     """
#     Compute clusters and return a list of lists of sha1
#     """
#     if not sha1s:
#         return []
#     values = await db.get_sha1_computed_values(sha1s)
#     clusters, distances = compute.make_clusters(values, method="kmeans", nb_clusters=sensibility)
#     return Clusters(clusters=clusters, distances=distances)
#
#
# async def get_similar_images(sha1s: list[str]):
#     vectors = [i.vector for i in await db.get_sha1_computed_values(sha1s)]
#     res = compute.get_similar_images(vectors)
#     return [img for img in res if img['sha1'] not in sha1s]
#
#
# async def add_property_to_images(property_id: int, sha1_list: list[str], value: JSON) -> str:
#     # first check that the property and the image exist:
#     images = await db.get_images(sha1s=sha1_list)
#     if await db.get_property_by_id(property_id) and images:
#         await db.set_property_values(image_ids=[img.id for img in images], property_id=property_id, value=value)
#         # check if a value already exists
#         # if await db.get_image_property(sha1, property_id):
#         #     await db.update_image_property(sha1, property_id, value)
#         # else:
#         #     await db.add_image_property(sha1, property_id, value)
#         return value
#     else:
#         raise HTTPException(status_code=400, detail="Trying to set a value on a non existent property or sha1")
#
#
# async def read_properties_file(data: pandas.DataFrame):
#     filenames, ids, sha1 = zip(*[(i.name, i.id, i.sha1) for i in await db.get_images()])
#
#     # if there are no duplicates them lets assume the props are sha1
#     if data.duplicated('key').sum() == 0:
#         prop_mode = "sha1"
#         matcher = dict(zip(filenames, sha1))
#     else:
#         prop_mode = "id"
#         matcher = dict(zip(filenames, ids))
#     data = data[data.key.isin(filenames)]
#     # data = data.drop_duplicates(subset='key', keep='first')
#     # add panoptic id or sha1 to the dataframe
#     data.loc[data.index, 'panoptic_id'] = data['key'].map(matcher)
#
#     # first, create new images where ids are the same
#     unique_ids = list(data.panoptic_id.unique())
#     clones_created = 0
#     # TODO: create clones all at once ? it's tricky with the ids to update
#     if prop_mode == "id":
#         for id_ in tqdm(unique_ids):
#             sub_data = data[data.panoptic_id == id_]
#             # create clones only if needed
#             if len(list(sub_data.panoptic_id)) == 0:
#                 continue
#             image = await db.get_images([id_])
#             image = image[0]
#             new_ids = await db.create_clones(image, sub_data.shape[0] - 1)
#             clones_created += len(new_ids)
#             data.loc[data.panoptic_id == id_, "panoptic_id"] = [image.id, *new_ids]
#         logging.getLogger('panoptic').info(f"created {clones_created} new images")
#     properties_to_create = data.columns.tolist()
#
#     # then for each property to create
#     for prop in properties_to_create:
#         if prop in ["key", "panoptic_id", "sha1"]:
#             continue
#
#         # create the property
#         prop_name, prop_type = prop.split('[')
#         prop_type = PropertyType(prop_type.split(']')[0])
#         property = await create_property(prop_name, prop_type, prop_mode)
#         # then get all possible values to insert
#         prop_values = list(data[prop].unique())
#
#         # if it's a tag property let's create the tags in the db
#         if property.id == PropertyType.tag or property.id == PropertyType.multi_tags:
#             tag_matcher = {}
#             # TODO: can we optimize to create all tags at once ?
#             for value in tqdm(prop_values):
#                 created_tags = []
#                 # if value is empty or empty quotes just create a "null" tag
#                 if pandas.isna(value) or str(value).strip() == "":
#                     value = "unknown"
#                 tags = [value] if property.id == PropertyType.tag else str(value).split(',')
#                 # if it's multi tag, assume tags are separated by a comma and create them separately
#                 for single_tag in tags:
#                     colors = range(11)
#                     color = random.randint(0, len(colors) - 1)
#                     parent_node = 0
#                     for child in single_tag.split('>'):
#                         tag = await create_tag(property.id, child, parent_node, color=color)
#                         parent_node = tag.id
#                         created_tags.append(tag.id)
#                 tag_matcher[value] = created_tags
#             # now change all values in the dataframe with the real value that we are going to insert
#             data.loc[data.index, prop] = data[prop].map(tag_matcher)
#         await db.set_multiple_property_values(property.id, list(data[prop]), list(data.panoptic_id))
#
#
# async def add_folder(folder):
#     found = await importer.import_folder(folder)
#     logging.info(f'found {found} images')
#     return folder
#
#
# new_images = []
#
#
# def get_new_images():
#     copy = [i for i in new_images]
#     new_images.clear()
#     return copy
#
#
# async def create_tag(property_id, value, parent_id, color: int) -> Tag:
#     existing_tag = await db.get_tag(property_id, value)
#     if existing_tag is not None:
#         if await db.tag_in_ancestors(existing_tag.id, parent_id):
#             raise HTTPException(status_code=400, detail="Adding a tag that is an ancestor of himself")
#         existing_tag.parents = list({*existing_tag.parents, parent_id})
#         await db.update_tag(existing_tag)
#         tag_id = existing_tag.id
#     else:
#         parents = [parent_id]
#         tag_id = await db.add_tag(property_id, value, json.dumps(parents), color)
#     return await db.get_tag_by_id(tag_id)
#
#
# async def tag_add_parent(tag_id, parent_id):
#     tag = await db.get_tag_by_id(tag_id)
#     parent = await db.get_tag_by_id(parent_id)
#     if tag is None:
#         raise HTTPException(status_code=400, detail=f"Tag: {tag_id} doesnt exist")
#     if parent is None:
#         raise HTTPException(status_code=400, detail=f"Parent Tag: {parent_id} doesnt exist")
#     if await db.tag_in_ancestors(tag.id, parent_id):
#         raise HTTPException(status_code=400, detail="Adding a tag that is an ancestor of himself")
#     tag.parents = list({*tag.parents, parent_id})
#     tag.color = parent.color
#     await db.update_tag(tag)
#     return tag
#
#
# async def update_tag(payload: UpdateTagPayload) -> Tag:
#     existing_tag = await db.get_tag_by_id(payload.id)
#     if not existing_tag:
#         raise HTTPException(status_code=400, detail="Trying to modify non existent tag")
#     # if await db.tag_in_ancestors(existing_tag.id, payload.parent_id):
#     #     raise HTTPException(status_code=400, detail="Adding a tag that is an ancestor of himself")
#     # change only fields of the tags that are set in the payload
#     # new_tag = existing_tag.copy(update=payload.dict(exclude_unset=True))
#     if payload.color is not None:
#         existing_tag.color = payload.color
#     if payload.value is not None:
#         existing_tag.value = payload.value
#
#     await db.update_tag(existing_tag)
#     return existing_tag
#
#
# async def delete_tag(tag_id: int) -> List[int]:
#     # first delete the tag
#     modified_tags = [await db.delete_tag_by_id(tag_id)]
#     # when deleting a tag, get all children of this tag
#     children = await db.get_tags_by_parent_id(tag_id)
#     # and remove parent ref of tag in children
#     [modified_tags.extend(await delete_tag_parent(c.id, tag_id)) for c in children]
#     # then get all images properties tagged with it
#     property_values = await db.get_property_values_with_tag(tag_id)
#     # and delete the tag ref from them
#     for data in property_values:
#         if tag_id in data.value:
#             data.value.remove(tag_id)
#             if data.image_id >= 0:
#                 await db.set_property_values(data.property_id, data.value, image_ids=[data.image_id])
#             else:
#                 await db.set_property_values(data.property_id, data.value, sha1s=[data.sha1])
#     return modified_tags
#
#
# async def delete_tag_parent(tag_id: int, parent_id: int) -> List[int]:
#     tag = await db.get_tag_by_id(tag_id)
#     tag.parents.remove(parent_id)
#     print(tag.parents)
#     color = random.randint(0, 11)
#     tag.color = color
#     print(color)
#     await db.update_tag(tag)
#     return tag
#
#
# async def get_tags(prop: str = None) -> Tags:
#     res = {}
#     tag_list = await db.get_tags(prop)
#     for tag in tag_list:
#         if tag.property_id not in res:
#             res[tag.property_id] = {}
#         res[tag.property_id][tag.id] = tag
#     return res



from __future__ import annotations

import math
import os
import pickle
from typing import List

import faiss
import numpy as np
from faiss import index_factory
from scipy.stats import hmean
from sklearn.cluster import DBSCAN, KMeans, estimate_bandwidth, MeanShift
from sklearn.metrics import silhouette_score
from sklearn.neighbors import KDTree
from panoptic.models import ComputedValue, Vector

from .transform import transformer
from .utils import load_similarity_tree


class SimilarityTreeWithLabel:
    def __init__(self, images: list[ComputedValue]):
        vectors, sha1_list = zip(*[(i.vector, i.sha1) for i in images])
        self.image_labels = sha1_list
        self.tree = KDTree(vectors)

    def query(self, image: np.ndarray, k=9999):
        dist, ind = self.tree.query(image.reshape(1, -1), k)
        indices = [x for x in ind[0]]
        distances = [x for x in dist[0]]
        return [{'sha1': self.image_labels[i], 'dist': float('%.2f' % (distances[index]))} for index, i in
                enumerate(indices)]


class SimilarityFaissWithLabel:
    def __init__(self, images: list[Vector]):
        vectors, sha1_list = zip(*[(i.data, i.sha1) for i in images])
        print(sha1_list)
        vectors = np.asarray(vectors)
        faiss.normalize_L2(vectors)
        self.image_labels = sha1_list
        # create the faiss index based on this post: https://anttihavanko.medium.com/building-image-search-with-openai-clip-5a1deaa7a6e2
        nb_vectors = vectors.shape[0]
        vector_size = vectors.shape[1]
        # reduce vector size only if we have more than 100k images otherwise it's not worth it since we lose accuracy
        if nb_vectors < 100000:
            index = faiss.IndexFlatIP(vector_size)
        else:
            cells = min(round(math.sqrt(nb_vectors)), int(nb_vectors / 39))
            index = index_factory(vector_size, f"IVF{cells},PQ16np")
            index.train(vectors)
            index.nprobe = 10
        self.tree = index
        self.tree.add(np.asarray(vectors))

    def query(self, image: np.ndarray, k=99999):
        faiss.normalize_L2(image)
        vector = image.reshape(1, -1)
        dist, ind = self.tree.search(vector, k)  # len(self.image_labels))
        indices = [x for x in ind[0]]
        distances = [x for x in dist[0]]
        return [{'sha1': self.image_labels[i], 'dist': float('%.2f' % (distances[index]))} for index, i in
                enumerate(indices)]


SIMILARITY_TREE: SimilarityTreeWithLabel | None = None


def reload_tree(path: str):
    global SIMILARITY_TREE
    SIMILARITY_TREE = load_similarity_tree(path)


def create_similarity_tree(path: str, images: list[ComputedValue]):
    tree = SimilarityTreeWithLabel(images)
    with open(os.path.join(path, 'tree.pkl'), 'wb') as f:
        pickle.dump(tree, f)
    global SIMILARITY_TREE
    SIMILARITY_TREE = tree


def create_similarity_tree_faiss(path: str, images: list[Vector]):
    tree = SimilarityFaissWithLabel(images)
    with open(os.path.join(path, 'tree_faiss.pkl'), 'wb') as f:
        pickle.dump(tree, f)
    global SIMILARITY_TREE
    SIMILARITY_TREE = tree


async def get_similar_images_from_text(input_text: str):
    if transformer.can_handle_text:
        vec = transformer.to_text_vector(input_text)
        return SIMILARITY_TREE.query(vec)


def get_similar_images(vectors: list[np.ndarray]):
    if not SIMILARITY_TREE:
        raise ValueError("Cannot compute image similarity since KDTree was not computed yet")
    vector = np.mean(vectors, axis=0)
    return SIMILARITY_TREE.query(np.asarray([vector]))


def make_clusters(vectors: List[Vector], *, method='kmeans', **kwargs) -> (list[list[str]], list[int]):
    res_clusters = []
    res_distances = []
    vectors, sha1 = zip(*[(i.data, i.sha1) for i in vectors])
    sha1 = np.asarray(sha1)
    clusters: np.ndarray
    distances: np.ndarray | None = None
    method = "faiss"
    # for now faiss method is better than the others, and probably this will be refactored with plugin system
    # so for now the matching is dead code
    clusters, distances = _make_clusters_faiss(vectors, **kwargs)
    # match method:
    #     case 'kmeans':
    #         clusters = _make_clusters_kmeans(vectors, **kwargs)
    #     case 'dbscan':
    #         clusters = _make_clusters_dbscan(vectors, **kwargs)
    #     case 'meanshift':
    #         clusters = _make_clusters_meanshift(vectors, **kwargs)
    #     case 'faiss':
    #         clusters, distances = _make_clusters_faiss(vectors, **kwargs)
    #     case other:
    #         return [[]]
    for cluster in list(set(clusters)):
        sha1_clusters = sha1[clusters == cluster]
        # sort by average_hash
        # sorted_cluster = [sha1 for _, sha1 in sorted(zip(ahashs_clusters, sha1_clusters))]
        if distances is not None:
            # clusters_distances = distances[clusters == cluster]
            # max_distance = np.max(clusters_distances)
            # clusters_distances / max_distance
            res_distances.append(hmean(distances[clusters == cluster]))
        res_clusters.append(list(sha1_clusters))
    # sort clusters by distances
    # TODO: trouver un meilleur indicateur que juste la moyenne des distances ?
    # TODO: virer les groupes avec une seule image ?
    sorted_clusters = [cluster for _, cluster in sorted(zip(res_distances, res_clusters))]
    return sorted_clusters, sorted(res_distances)


def _make_clusters_dbscan(vectors, eps=3, *args, **kwargs) -> np.ndarray:
    clusters = DBSCAN(eps=eps).fit(np.asarray(vectors))
    return clusters.labels_


def _make_clusters_kmeans(vectors, nb_clusters=6, *args, **kwargs) -> np.ndarray:
    clusters = KMeans(n_clusters=int(nb_clusters)).fit_predict(vectors)
    return clusters


def _make_clusters_faiss(vectors, nb_clusters=6, *args, **kwargs) -> (np.ndarray, np.ndarray):
    def _make_single_kmean(vectors, nb_clusters):
        kmean = faiss.Kmeans(vectors.shape[1], nb_clusters, niter=20, verbose=False)
        kmean.train(vectors)
        return kmean.index.search(vectors, 1)

    vectors = np.asarray(vectors)
    if nb_clusters == 0:
        k_silhouettes = []
        for k in range(3, 30):
            distances, indices = _make_single_kmean(vectors, k)
            indices = indices.flatten()
            k_silhouettes.append(silhouette_score(vectors, indices))
        nb_clusters = int(np.argmax(k_silhouettes)) + 3
    distances, indices = _make_single_kmean(vectors, nb_clusters)
    return indices.flatten(), distances.flatten()


def _make_clusters_meanshift(vectors, *args, **kwargs) -> np.ndarray:
    bandwidth = estimate_bandwidth(vectors, quantile=0.2, n_samples=500)
    clusters = MeanShift(bandwidth=bandwidth, bin_seeding=True).fit(np.asarray(vectors))
    return clusters.labels_



import asyncio
from concurrent.futures import Executor


class Task:
    """
    Base class for a Task managed by the TaskQueue.
    To use the task class outside the TaskQueue you need to set_executor manually
    """
    def __init__(self, priority=False):
        self._executor: Executor | None = None
        self.has_priority = priority
        self.name = type(self).__name__

    def set_executor(self, executor: Executor):
        self._executor = executor

    def get_id(self):
        return str(self.__class__)

    async def _async(self, function, *args):
        """
        Make function awaitable and execute in Executor
        """
        return await asyncio.wrap_future(self._executor.submit(function, *args))

    async def run(self):
        raise NotImplementedError()

    async def run_if_last(self):
        pass



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from panoptic.models import InstancePropertyValue, Tag


@dataclass(slots=True)
class Group:
    ids: list[int]
    score: int = None
    meta: Any = None


@dataclass(slots=True)
class GroupResult:
    groups: list[Group]


@dataclass(slots=True)
class InstanceMatch:
    id: int
    score: float


@dataclass(slots=True)
class SearchResult:
    matches: list[InstanceMatch]


@dataclass
class ActionError:
    type: str
    data: Any


@dataclass
class ActionResult:
    action: str
    result: Any
    errors: list[ActionError] = None


@dataclass(slots=True)
class DeleteTagResult:
    tag_id: int
    updated_values: list[InstancePropertyValue]
    updated_tags: list[Tag]



import glob
import os
import pathlib

import psutil
from fastapi import APIRouter
from pydantic import BaseModel

from panoptic.core.panoptic import Panoptic
from panoptic.models import PathRequest

selection_router = APIRouter()

panoptic: Panoptic | None = None


def set_panoptic(pano: Panoptic):
    global panoptic
    panoptic = pano


class ProjectRequest(BaseModel):
    path: str
    name: str


@selection_router.get("/status")
async def get_status_route():
    return {
        'isLoaded': panoptic.is_loaded(),
        'selectedProject': panoptic.project_id,
        'projects': panoptic.data.projects
    }


@selection_router.post("/load")
async def load_project_route(path: PathRequest):
    await panoptic.load_project(path.path)
    return await get_status_route()


@selection_router.post("/close")
async def close_project():
    await panoptic.close_project()
    return await get_status_route()


@selection_router.post("/delete_project")
async def delete_project_route(req: PathRequest):
    panoptic.remove_project(req.path)
    return await get_status_route()


@selection_router.post("/create_project")
async def create_project_route(req: ProjectRequest):
    await panoptic.create_project(req.name, req.path)
    return await get_status_route()


@selection_router.post("/import_project")
async def import_project_route(req: PathRequest):
    await panoptic.import_project(req.path)
    return await get_status_route()


@selection_router.get("/filesystem/ls/{path:path}")
def api(path: str = ""):
    return list_contents(path)


@selection_router.get('/filesystem/info')
def filesystem_info_route():
    return list_index()


@selection_router.get('/plugins')
async def get_plugins_route():
    return panoptic.get_plugin_paths()


@selection_router.post('/plugins')
async def add_plugins_route(path: PathRequest):
    return panoptic.add_plugin_path(path.path)


@selection_router.delete('/plugins')
async def del_plugins_route(path: str):
    return panoptic.del_plugin_path(path)


def images_in_folder(folder_path):
    types = ('*.jpg', '*.jpeg', '*.png', '*.gif', '*.bmp')  # the tuple of file types
    image_files = []
    for type_ in types:
        image_files.extend(glob.glob(os.path.join(folder_path, type_)))

    return image_files


def list_contents(full_path: str = '/'):
    paths = [full_path + '/' + p if full_path != '/' else full_path + p for p in os.listdir(full_path)]
    directories = [p for p in paths if os.path.isdir(p)]
    directories = [{
        'path': p,
        'name': pathlib.Path(p).name,
        'images': len(images_in_folder(p)),
        'isProject': os.path.exists(os.path.join(p, 'panoptic.db'))
    } for p in directories]
    images = images_in_folder(full_path)

    return {'images': images[0:40], 'directories': directories}


def list_disk():
    files = []
    partitions = psutil.disk_partitions()
    partitions = [p for p in partitions if not p.mountpoint.startswith("/System")]
    for partition in partitions:
        files.append({
            'path': partition.mountpoint,
            'name': pathlib.Path(partition.mountpoint).name,
            'images': len(images_in_folder(partition.mountpoint))
        })
    return files


def list_index():
    mounted = []

    partitions = psutil.disk_partitions()
    partitions = [p for p in partitions if not p.mountpoint.startswith("/System")]
    for partition in partitions:
        mounted.append({
            'path': partition.mountpoint,
            'name': partition.mountpoint,
            'images': len(images_in_folder(partition.mountpoint))
        })

    files = [{
        'path': pathlib.Path.home(),
        'name': 'Home',
        'images': len(images_in_folder(pathlib.Path.home()))
    }]

    home_files = list_contents(str(pathlib.Path.home()))['directories']
    home_files = [f for f in home_files if f['name'] in ['Documents', 'Downloads', 'Desktop', 'Images', 'Pictures']]
    files.extend(home_files)
    return {'partitions': mounted, 'fast': files}



from __future__ import annotations
import hashlib
import os
from typing import TYPE_CHECKING

import numpy
from PIL import Image
from imagehash import MeanFunc, ImageHash, ANTIALIAS

if TYPE_CHECKING:
    from panoptic.core.project.project import Project

from panoptic.core.task.task import Task


def average_hash(image: Image.Image, hash_size: int = 8, mean: MeanFunc = numpy.mean) -> ImageHash:
    """
	Average Hash computation

	Implementation follows http://www.hackerfactor.com/blog/index.php?/archives/432-Looks-Like-It.html

	Step by step explanation: https://web.archive.org/web/20171112054354/https://www.safaribooksonline.com/blog/2013/11/26/image-hashing-with-python/ # noqa: E501

	@image must be a PIL instance.
	@mean how to determine the average luminescence. can try numpy.median instead.
    """
    if hash_size < 2:
        raise ValueError('Hash size must be greater than or equal to 2')

    # reduce size and complexity, then covert to grayscale
    image = image.convert('L').resize((hash_size, hash_size), ANTIALIAS)

    # find average pixel value; 'pixels' is an array of the pixel values, ranging from 0 (black) to 255 (white)
    pixels = numpy.asarray(image)
    avg = mean(pixels)

    # create string of bits
    diff = pixels > avg
    # make a hash
    return ImageHash(diff)


class ImportInstanceTask(Task):
    def __init__(self, project: Project, file: str, folder_id: int):
        super().__init__(priority=True)
        self.project = project
        self.db = project.db
        self.file = file
        self.folder_id = folder_id
        self.name = 'Import Image'

    async def run(self):
        name = self.file.split(os.sep)[-1]
        extension = name.split('.')[-1]
        folder_id = self.folder_id

        raw_db = self.db.get_raw_db()

        db_image = await raw_db.has_file(folder_id, name, extension)
        if db_image:
            self.db.on_import_instance.emit(db_image)
            return db_image

        sha1, url, width, height, ahash = await self._async(self._import_image, self.file, raw_db.get_project_path())

        image = await self.db.add_instance(folder_id, name, extension, sha1, url, width, height, str(ahash))
        # print(f'imported image: {image.id} : {image.sha1}')
        return image

    async def run_if_last(self):
        self.project.ui.update_counter.image += 1

    @staticmethod
    def _import_image(file_path, project_path: str):
        image = Image.open(file_path)
        width, height = image.size
        sha1_hash = hashlib.sha1(image.tobytes()).hexdigest()
        # TODO: gérer l'url statique quand on sera en mode serveur
        # url = os.path.join('/static/' + global_file_path.split(os.getenv('PANOPTIC_ROOT'))[1].replace('\\', '/'))
        if not os.path.exists(os.path.join(project_path, "mini")):
            os.mkdir(os.path.join(project_path, "mini"))
        url = file_path
        image = image.convert('RGB')
        mini = image.copy()
        mini.thumbnail(size=(200, 200))
        mini.save(os.path.join(project_path, "mini", sha1_hash + '.jpeg'), optimize=True, quality=30)

        ahash = average_hash(image)

        del image
        del mini
        # gc.collect()

        return sha1_hash, url, width, height, ahash



import atexit
import os
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import List

from showinfm import show_in_file_manager

from panoptic.core.db.db_connection import DbConnection
from panoptic.core.exporter import Exporter
from panoptic.core.project.project_actions import ProjectActions
from panoptic.core.project.project_db import ProjectDb
from panoptic.core.project.project_events import ProjectEvents
from panoptic.core.project.project_ui import ProjectUi
from panoptic.core.task.import_image_task import ImportInstanceTask
from panoptic.core.task.load_plugin_task import LoadPluginTask
from panoptic.core.task.task_queue import TaskQueue
from panoptic.core.importer import Importer
from panoptic.models import StatusUpdate, PluginDefaultParams, ActionParam
from panoptic.plugin import Plugin

nb_workers = 4


def get_executor():
    executor = ThreadPoolExecutor(max_workers=nb_workers)
    atexit.register(executor.shutdown)
    return executor


class Project:
    def __init__(self, folder_path: str, plugins: List[str]):
        self.executor = get_executor()
        self.is_loaded = False
        self.base_path = folder_path
        self.db: ProjectDb | None = None
        self.ui: ProjectUi | None = None
        self.on = ProjectEvents()
        self.action = ProjectActions()
        self.task_queue = TaskQueue(self.executor, num_workers=4)
        self.importer = Importer(project=self)
        self.exporter = Exporter(project=self)

        self.plugin_loaded = False
        self.plugins: List[Plugin] = []
        self.plugin_paths = plugins

    async def start(self):
        conn = DbConnection(self.base_path)
        await conn.start()
        self.db = ProjectDb(conn)
        self.ui = ProjectUi(self.db.get_raw_db())

        self.db.on_import_instance.redirect(self.on.import_instance)

        for plugin_path in self.plugin_paths:
            task = LoadPluginTask(self, plugin_path)
            self.task_queue.add_task(task)

        self.is_loaded = True

    async def redirect_on_import(self, x):
        self.on.import_instance.emit(x)

    def get_status_update(self) -> StatusUpdate:
        res = StatusUpdate(update=self.ui.update_counter)
        res.tasks = self.task_queue.get_task_states()
        res.plugin_loaded = self.plugin_loaded
        return res

    async def close(self):
        self.is_loaded = False
        self.base_path = ''
        await self.db.close()

    async def export_data(self, name: str = None, ids: [int] = None, properties: [int] = None,
                          copy_images: bool = False):
        export_path = await self.exporter.export_data(name=name, path=self.base_path, instance_ids=ids,
                                                      properties=properties,
                                                      copy_images=copy_images)
        # export_path = "lala"
        show_in_file_manager(export_path)
        return export_path

    async def plugins_info(self):
        return [await p.get_description() for p in self.plugins]

    async def import_folder(self, folder: str):
        all_files = [os.path.join(path, name) for path, subdirs, files in os.walk(folder) for name in files]
        all_images = [i for i in all_files if
                      i.lower().endswith('.png') or i.lower().endswith('.jpg') or i.lower().endswith('.jpeg')]

        folder_node, file_to_folder_id = await self._compute_folder_structure(folder, all_images)

        tasks = [ImportInstanceTask(project=self, file=file, folder_id=file_to_folder_id[file])
                 for file in all_images]
        [self.task_queue.add_task(t) for t in tasks]

    async def _compute_folder_structure(self, root_path, all_files: List[str]):
        offset = len(root_path)
        root, root_name = os.path.split(root_path)
        root_folder = await self.db.add_folder(root_path, root_name)
        file_to_folder_id = {}
        for file in all_files:
            path, name = os.path.split(file)
            if offset == len(path):
                file_to_folder_id[file] = root_folder.id
                continue
            path = path[offset + 1:]
            parts = Path(path).parts
            current_folder = root_folder
            for part in parts:
                if part not in current_folder.children:
                    child = await self.db.add_folder(current_folder.path + '/' + part, part, current_folder.id)
                    current_folder.children[part] = child
                else:
                    child = current_folder.children[part]
                file_to_folder_id[file] = child.id
                current_folder = child
        return root_folder, file_to_folder_id

    async def set_plugin_default_params(self, params: PluginDefaultParams):
        await self.db.set_plugin_default_params(params)
        plugin = [p for p in self.plugins if p.name == params.name][0]
        plugin.update_default_values(params)
        return params

    async def set_action_updates(self, updates: List[ActionParam]):
        self.action.set_action_functions(updates)
        for action in updates:
            # make sure we don't override other ActionParams
            if action.name in self.action.actions:
                await self.db.get_raw_db().set_action_param(action)

    async def update_actions_from_db(self):
        db_actions = await self.db.get_raw_db().get_action_params()
        db_actions = [a for a in db_actions if a.value]
        self.action.set_action_functions(db_actions)



# Connexion à la DB SQLite
from __future__ import annotations

import json
from typing import List, Any

import numpy as np
from pypika import Table, Parameter, PostgreSQLQuery

from panoptic.core.db.db_connection import DbConnection
from panoptic.core.db.utils import auto_dict
from panoptic.models import Instance, ComputedValue, Vector, PluginDefaultParams, ActionParam, \
    VectorDescription, InstancePropertyValue, ImagePropertyValue
from panoptic.models import Tag, Property, Folder, Tab

Query = PostgreSQLQuery


class Db:
    def __init__(self, conn: DbConnection):
        if not conn.is_loaded:
            raise Exception('DbConnection is not started. Execute await conn.start() before')
        self.conn = conn

    async def close(self):
        await self.conn.conn.close()

    def get_project_path(self):
        return self.conn.folder_path

    # =====================================================
    # =================== Properties ======================
    # =====================================================

    async def add_property(self, name: str, property_type: str, mode: str) -> Property:
        query = 'INSERT INTO properties (name, type, mode) VALUES (?, ?, ?) on conflict do nothing'
        cursor = await self.conn.execute_query(query, (name, property_type, mode))
        prop = Property(id=cursor.lastrowid, name=name, type=property_type, mode=mode)
        return prop

    async def import_property(self, id_: int, name: str, property_type: str, mode: str) -> Property:
        query = 'INSERT INTO properties (id, name, type, mode) VALUES (?, ?, ?, ?) on conflict do nothing'
        cursor = await self.conn.execute_query(query, (id_, name, property_type, mode))
        prop = Property(id=cursor.lastrowid, name=name, type=property_type, mode=mode)
        return prop

    async def get_properties(self) -> list[Property]:
        query = "SELECT * from properties"
        cursor = await self.conn.execute_query(query)
        return [Property(**auto_dict(row, cursor)) for row in await cursor.fetchall()]

    async def get_property(self, property_id) -> [Property | None]:
        query = """
                SELECT *
                FROM properties
                WHERE id = ?
            """
        cursor = await self.conn.execute_query(query, (property_id,))
        row = await cursor.fetchone()
        if row:
            return Property(**auto_dict(row, cursor))
        else:
            return None

    async def update_property(self, new_property: Property):
        query = "UPDATE properties SET name = ?, id = ? WHERE id = ?"
        await self.conn.execute_query(query, (new_property.name, new_property.type.value, new_property.id))

    async def delete_property(self, property_id):
        query = "DELETE from properties WHERE id = ?"
        await self.conn.execute_query(query, (property_id,))

    # =====================================================
    # =============== Property Values =====================
    # =====================================================

    async def get_instance_property_values(self, property_ids: List[int] = None, instance_ids: list[int] = None) \
            -> list[InstancePropertyValue]:
        values = Table('instance_property_values')
        query = Query.from_(values).select('*')

        if property_ids:
            query = query.where(values.property_id.isin(property_ids))

        if instance_ids:
            query = query.where(values.instance_id.isin(instance_ids))

        cursor = await self.conn.execute_query(query.get_sql())
        res = [InstancePropertyValue(**auto_dict(instance, cursor)) for instance in await cursor.fetchall()]
        return res

    async def get_image_property_values(self, property_ids: List[int] = None, sha1s: list[str] = None) \
            -> list[ImagePropertyValue]:
        values = Table('image_property_values')
        query = Query.from_(values).select('*')

        if property_ids:
            query = query.where(values.property_id.isin(property_ids))

        if sha1s:
            query = query.where(values.sha1.isin(sha1s))

        cursor = await self.conn.execute_query(query.get_sql())
        res = [ImagePropertyValue(**auto_dict(image, cursor)) for image in await cursor.fetchall()]
        return res

    async def set_instance_property_value(self, property_id: int, instance_ids: List[int], value: Any):
        json_value = json.dumps(value)
        t = Table('instance_property_values')
        query = Query.into(t).columns('property_id', 'instance_id', 'value')
        query = query.insert(*[(property_id, iid, json_value) for iid in instance_ids])
        query = query.get_sql() + ' ON CONFLICT (property_id, instance_id) DO UPDATE SET value=excluded.value'
        await self.conn.execute_query(query)
        return value

    async def set_image_property_value(self, property_id: int, sha1s: List[str], value: Any):
        json_value = json.dumps(value)

        t = Table('image_property_values')
        query = Query.into(t).columns('property_id', 'sha1', 'value')
        query = query.insert(*[(property_id, sha1, json_value) for sha1 in sha1s])
        query = query.get_sql() + ' ON CONFLICT (property_id, sha1) DO UPDATE SET value=excluded.value'
        await self.conn.execute_query(query)
        return value

    async def set_instance_property_array_values(self, property_id: int, instance_ids: list[int], values: list[Any]):
        """
            Allows to set different values to each instance.
        """
        t = Table('instance_property_values')
        query = Query.into(t).columns('property_id', 'instance_id', 'value')
        query = query.insert((property_id, Parameter('?'), Parameter('?')))
        query = query.get_sql() + ' ON CONFLICT (property_id, instance_id) DO UPDATE SET value=excluded.value'
        await self.conn.execute_query_many(query, [(idd, json.dumps(value)) for idd, value in
                                                   zip(instance_ids, values)])

    async def set_image_property_array_values(self, property_id: int, sha1s: list[str], values: list[Any]):
        """
            Allows to set different values to each image.
        """
        t = Table('image_property_values')
        query = Query.into(t).columns('property_id', 'sha1', 'value')
        query = query.insert((property_id, Parameter('?'), Parameter('?')))
        query = query.get_sql() + ' ON CONFLICT (property_id, sha1) DO UPDATE SET value=excluded.value'
        await self.conn.execute_query_many(query, [(sha1, json.dumps(value)) for sha1, value in
                                                   zip(sha1s, values)])

    async def delete_instance_property_value(self, property_id: int, instance_ids: list[int]):
        t = Table('instance_property_values')
        query = Query.from_(t).delete().where(t.property_id == property_id)
        query = query.where(t.instance_id.isin(instance_ids))
        await self.conn.execute_query(query.get_sql())
        return True

    async def delete_image_property_value(self, property_id: int, sha1s: list[str]):
        t = Table('image_property_values')
        query = Query.from_(t).delete().where(t.property_id == property_id)
        query = query.where(t.sha1.isin(sha1s))
        await self.conn.execute_query(query.get_sql())
        return True

    # =====================================================
    # ====================== Tag ==========================
    # =====================================================

    async def add_tag(self, property_id: int, value: str, parents: str, color: int):
        query = "INSERT INTO tags (property_id, value, parents, color) VALUES (?, ?, ?, ?)"
        cursor = await self.conn.execute_query(query, (property_id, value, parents, color))
        return cursor.lastrowid

    async def import_tag(self, id_: int, property_id: int, value: str, parents: str, color: int):
        query = "INSERT INTO tags (id, property_id, value, parents, color) VALUES (?, ?, ?, ?, ?)"
        cursor = await self.conn.execute_query(query, (id_, property_id, value, parents, color))
        return cursor.lastrowid

    async def get_tag_by_id(self, tag_id):
        query = "SELECT * FROM tags WHERE id = ?"
        cursor = await self.conn.execute_query(query, (tag_id,))
        row = await cursor.fetchone()
        if not row:
            return
        return Tag(**auto_dict(row, cursor))

    async def get_tag(self, property_id, value):
        query = "SELECT * FROM tags WHERE property_id = ? AND value = ?"
        cursor = await self.conn.execute_query(query, (property_id, value))
        row = await cursor.fetchone()
        if not row:
            return
        return Tag(**auto_dict(row, cursor))

    async def update_tag(self, tag: Tag):
        query = "UPDATE tags SET parents = ?, value = ?, color = ? WHERE id = ?"
        await self.conn.execute_query(query, (json.dumps(tag.parents), tag.value, tag.color, tag.id))

    async def get_tag_ancestors(self, tag: Tag, acc=None):
        if not acc:
            acc = []
        if tag.parents == [0]:
            return list({*tag.parents, *acc})
        else:
            acc = acc + tag.parents
            for parent in tag.parents:
                if parent == 0:
                    continue
                parent_tag = await self.get_tag_by_id(parent)
                return await self.get_tag_ancestors(parent_tag, acc)

    async def get_tags(self, prop: int) -> list[Tag]:
        query = "SELECT * FROM tags "
        params = None
        if prop:
            query += "WHERE property_id = ?"
            params = (prop,)
        cursor = await self.conn.execute_query(query, params)
        return [Tag(**auto_dict(row, cursor)) for row in await cursor.fetchall()]

    async def delete_tag_by_id(self, tag_id: int) -> int:
        query = "DELETE FROM tags WHERE id = ?"
        await self.conn.execute_query(query, (tag_id,))
        return tag_id

    async def get_tags_by_parent_id(self, parent_id: int):
        query = "SELECT tags.* FROM tags, json_each(tags.parents) WHERE json_each.value = ?"
        cursor = await self.conn.execute_query(query, (parent_id,))
        return [Tag(**auto_dict(row, cursor)) for row in await cursor.fetchall()]

    async def tag_in_ancestors(self, tag_id, parent_id) -> bool:
        if not parent_id:
            return False
        else:
            parent = await self.get_tag_by_id(parent_id)
            ancestors = await self.get_tag_ancestors(parent)
            if tag_id in ancestors:
                return True
        return False

    # =====================================================
    # ================= Instances =========================
    # =====================================================

    async def add_instance(self, folder_id: int, name: str, extension: str, sha1: str, url: str, width: int,
                           height: int,
                           ahash: str):
        table = Table('instances')
        query = Query.into(table).columns(
            'folder_id',
            'name',
            'extension',
            'sha1',
            'url',
            'width',
            'height',
            'ahash').insert((
            folder_id,
            name,
            extension,
            sha1,
            url,
            width,
            height,
            ahash))
        cursor = await self.conn.execute_query(query.get_sql())
        id_ = cursor.lastrowid

        return Instance(id=id_, folder_id=folder_id, name=name, extension=extension, sha1=sha1, url=url, width=width,
                        height=height)

    async def import_instance(self, id_: int, folder_id: int, name: str, extension: str, sha1: str, url: str, width: int,
                           height: int,
                           ahash: str):
        table = Table('instances')
        query = Query.into(table).columns(
            'id',
            'folder_id',
            'name',
            'extension',
            'sha1',
            'url',
            'width',
            'height',
            'ahash').insert((
            id_,
            folder_id,
            name,
            extension,
            sha1,
            url,
            width,
            height,
            ahash))
        cursor = await self.conn.execute_query(query.get_sql())

        return Instance(id=id_, folder_id=folder_id, name=name, extension=extension, sha1=sha1, url=url, width=width,
                        height=height)

    async def clone_instance(self, i: Instance):
        return await self.add_instance(i.folder_id, i.name, i.extension, i.sha1, i.url, i.width, i.height, i.ahash)

    async def get_instances(self, ids: List[int] = None, sha1s: List[str] = None):
        img_table = Table('instances')
        query = Query.from_(img_table).select('*')

        if ids:
            query = query.where(img_table.id.isin(ids))
        if sha1s:
            query = query.where(img_table.sha1.isin(sha1s))

        cursor = await self.conn.execute_query(query.get_sql())
        instance = [Instance(*instance) for instance in await cursor.fetchall()]
        return instance

    async def has_file(self, folder_id, name, extension):
        table = Table('instances')
        query = Query.from_(table).select('*').where(table.folder_id == folder_id)
        query = query.where(table.name == name).where(table.extension == extension).limit(1)

        cursor = await self.conn.execute_query(query.get_sql())
        res = await cursor.fetchone()
        if res:
            return Instance(*res)
        return False

    # =====================================================
    # ================== Folders ==========================
    # =====================================================

    async def add_folder(self, path: str, name: str, parent: int = None):
        query = 'INSERT INTO folders (path, name, parent) VALUES (?, ?, ?) ON CONFLICT(path) DO NOTHING'
        await self.conn.execute_query(query, (path, name, parent))
        cursor = await self.conn.execute_query('SELECT * FROM folders WHERE path = ?', (path,))
        row = await cursor.fetchone()
        folder = Folder(**auto_dict(row, cursor))
        return folder

    async def import_folder(self, id_: int, path: str, name: str, parent: int = None):
        query = 'INSERT INTO folders (id, path, name, parent) VALUES (?, ?, ?, ?) ON CONFLICT(path) DO NOTHING'
        await self.conn.execute_query(query, (id_, path, name, parent))
        cursor = await self.conn.execute_query('SELECT * FROM folders WHERE path = ?', (path,))
        row = await cursor.fetchone()
        folder = Folder(**auto_dict(row, cursor))
        return folder

    async def get_folders(self) -> List[Folder]:
        query = "SELECT * from folders"
        cursor = await self.conn.execute_query(query)
        return [Folder(**auto_dict(row, cursor)) for row in await cursor.fetchall()]

    async def get_folder(self, folder_id: int):
        table = Table('folders')
        query = Query.from_(table).select('*').where(table.id == folder_id)
        cursor = await self.conn.execute_query(query.get_sql())
        row = await cursor.fetchone()

        return Folder(**auto_dict(row, cursor))

    async def get_folder_by_path(self, path: str):
        table = Table('folders')
        query = Query.from_(table).select('*').where(table.path == path)
        cursor = await self.conn.execute_query(query.get_sql())
        row = await cursor.fetchone()
        if row:
            return Folder(**auto_dict(row, cursor))
        return None

    async def delete_folder(self, folder_id: int):
        query = "DELETE from folders WHERE id = ?"
        await self.conn.execute_query(query, (folder_id,))
        return folder_id

    # =====================================================
    # ================== UI TABS ==========================
    # =====================================================

    async def add_tab(self, data):
        query = 'INSERT INTO tabs (data) VALUES (?)'
        cursor = await self.conn.execute_query(query, (json.dumps(data),))
        row = await cursor.fetchone()
        tab = Tab(id=cursor.lastrowid, data=data)
        return tab

    async def get_tabs(self) -> List[Tab]:
        query = "SELECT * from tabs"
        cursor = await self.conn.execute_query(query)
        return [Tab(**auto_dict(row, cursor)) for row in await cursor.fetchall()]

    async def update_tab(self, tab: Tab):
        query = "UPDATE tabs SET data = ? WHERE id = ?"
        await self.conn.execute_query(query, (json.dumps(tab.data), tab.id))

    async def delete_tab(self, tab_id: int):
        query = "DELETE from tabs WHERE id = ?"
        await self.conn.execute_query(query, (tab_id,))
        return tab_id

    # =====================================================
    # ================== Vectors ==========================
    # =====================================================

    async def add_vector(self, vector: Vector):
        query = f"""
            INSERT OR REPLACE INTO vectors (source, type, sha1, data)
            VALUES (?, ?, ?, ?);
        """
        await self.conn.execute_query(query, (vector.source, vector.type, vector.sha1, vector.data))
        return vector

    async def get_vectors(self, source: str, type_: str, sha1s: List[str] = None):
        t = Table('vectors')
        query = Query.from_(t).select('source', 'type', 'sha1', 'data')
        query = query.where(t.source == source).where(t.type == type_)
        if sha1s:
            query = query.where(t.sha1.isin(sha1s))
        cursor = await self.conn.execute_query(query.get_sql())
        rows = await cursor.fetchall()
        res = [Vector(*row) for row in rows]
        return res

    async def vector_exist(self, source: str, type_: str, sha1: str):
        t = Table('vectors')
        query = Query.from_(t).select('source', 'type', 'sha1')
        query = query.where(t.source == source).where(t.type == type_)
        query = query.where(t.sha1 == sha1)
        cursor = await self.conn.execute_query(query.get_sql())
        rows = await cursor.fetchall()
        if rows:
            return True
        return False

    async def get_vector_descriptions(self):
        query = """ 
        SELECT source, type, count(sha1) as count
        FROM vectors
        GROUP BY source, type;
        """
        cursor = await self.conn.execute_query(query)
        rows = await cursor.fetchall()
        if rows:
            return [VectorDescription(**auto_dict(r, cursor)) for r in rows]
        return []

    # =====================================================
    # ================== Plugins ==========================
    # =====================================================

    async def get_plugin_default_params(self, plugin_name: str):
        t = Table('plugin_defaults')
        query = Query.from_(t).select('name', 'base', 'functions')
        query = query.where(t.name == plugin_name)
        cursor = await self.conn.execute_query(query.get_sql())
        row = await cursor.fetchone()
        if row:
            return PluginDefaultParams(**auto_dict(row, cursor))
        return PluginDefaultParams(name=plugin_name)

    async def set_plugin_default_params(self, params: PluginDefaultParams):
        query = f"""
                    INSERT OR REPLACE INTO plugin_defaults (name, base, functions)
                    VALUES (?, ?, ?);
                """
        await self.conn.execute_query(query, (params.name, json.dumps(params.base), json.dumps(params.functions)))
        return params

    # =====================================================
    # ================== Actions ==========================
    # =====================================================

    async def get_action_params(self):
        query = "SELECT * FROM action_params"
        cursor = await self.conn.execute_query(query)
        rows = await cursor.fetchall()
        if rows:
            return [ActionParam(**auto_dict(r, cursor)) for r in rows]
        return []

    async def get_action_param(self, name: str):
        query = "SELECT * FROM action_params WHERE name=?"
        cursor = await self.conn.execute_query(query, (name,))
        row = await cursor.fetchone()
        if row:
            return ActionParam(**auto_dict(row, cursor))
        return None

    async def set_action_param(self, update: ActionParam):
        print('update', update)
        query = f"""
                    INSERT OR REPLACE INTO action_params (name, value)
                    VALUES (?, ?);
                """
        await self.conn.execute_query(query, (update.name, update.value))
        return update



import logging
import os
import sqlite3

import aiosqlite
import numpy as np

from panoptic.core.db.create import tables, DB_VERSION, software_db_version

aiosqlite.register_adapter(np.array, lambda arr: arr.tobytes())
aiosqlite.register_converter("array", lambda arr: np.frombuffer(arr, dtype='float32'))


class DbConnection:
    def __init__(self, path):
        self.folder_path = path
        self.db_path = os.path.join(path, "panoptic.db")
        self.is_loaded = False
        self.conn: aiosqlite.Connection | None = None

    async def start(self):
        """
        Startup function should be called first
        Cannot be done in __init__ because of async
        """
        # create connection
        self.conn = await aiosqlite.connect(self.db_path, detect_types=sqlite3.PARSE_DECLTYPES)
        # use foreign_keys checks
        async with self.conn.executescript('PRAGMA foreign_keys = 1') as cursor:
            await self.conn.commit()

        # check if the db is empty or has missing tables
        for table_name, create_query in tables.items():
            # Check if the table exists
            if not await self._table_exists(table_name):
                # If not, create the table
                async with self.conn.executescript(create_query) as cursor:
                    await self.conn.commit()
                print(f"Table '{table_name}' created.")

        # check if db version is the same as the software version
        db_version = await self.get_param(DB_VERSION)
        if int(db_version) != software_db_version:
            logging.warning(f'DB Version ({db_version}) doesnt match Software DB Version ({software_db_version})')
        self.is_loaded = True

    async def execute_query(self, query: str, parameters: tuple = None):
        cursor = await self.conn.cursor()
        if parameters:
            await cursor.execute(query, parameters)
        else:
            await cursor.execute(query)
        await self.conn.commit()
        return cursor

    async def execute_query_many(self, query, data: list):
        cursor = await self.conn.cursor()
        await cursor.executemany(query, data)
        await self.conn.commit()
        return cursor

    async def _table_exists(self, table_name: str):
        # Use the sqlite_master table to check if the table exists
        query = f"SELECT name FROM sqlite_master WHERE type='table' AND name='{table_name}'"
        cursor = await self.execute_query(query)
        result = await cursor.fetchone()

        return result is not None

    async def get_param(self, key: str) -> str | None:
        query = f'SELECT value FROM panoptic WHERE key="{key}";'
        cursor = await self.execute_query(query)
        row = await cursor.fetchone()
        if row:
            return row[0]
        return None

    async def set_param(self, key: str, value: str):
        query = f"""
                INSERT OR REPLACE INTO panoptic (key, value)
                VALUES ('{key}', '{value}');
        """
        cursor = await self.execute_query(query)
        return cursor



import json
from json import JSONDecodeError


def auto_dict(row, cursor):
    return {key: decode_if_json(value) for key, value in zip([c[0] for c in cursor.description], row)}


def decode_if_json(value):
    try:
        return json.loads(value)
    except (TypeError, JSONDecodeError, UnicodeDecodeError):
        return value



import os
import pickle

from tqdm import tqdm

from plugins.FaissPlugin.compute import create_pca, to_pca, create_similarity_tree, can_compute_pca
from panoptic.core.db import db
from panoptic.models import ComputedValue


async def compute_all_pca(project_path: str, force=False):
    # await db_utils.init()
    all_images: list[ComputedValue] = await db.get_sha1_computed_values()
    if not force and not can_compute_pca(len(all_images), all_images[0].vector):
        pass
    else:
        vectors = [i.vector for i in all_images]
        with open(os.path.join(project_path, 'vectors.pkl'), 'wb') as f:
            pickle.dump(vectors, f)
        create_pca(vectors, project_path)
        for i, v in tqdm(zip(all_images, vectors)):
            pca_vec = to_pca(v)
            await db.set_computed_value(i.sha1, i.ahash, pca_vec)
    all_images_pca: list[ComputedValue] = await db.get_sha1_computed_values()
    create_similarity_tree(project_path, all_images_pca)
    await db.vacuum()
    return


import os
import sys

# import cv2
import numpy as np
import pandas as pd
import pytesseract

os.environ['TESSDATA_PREFIX'] = os.getenv('TESSDATA_PREFIX', os.sep.join([*sys.executable.split(os.sep)[:-1], 'share', 'tessdata']))

tess_config = "-c tessedit_char_blacklist='+=*;)(][\{\}@/|_^\\#~&'"

#
# def full_ocr(image: Instance):
#     img = np.array(image)
#
#     # create list of combinations (preprocessing + language) to try
#     operations = list(itertools.product([_no_preprocess, _bilateral], ['eng', 'fra']))
#
#     # run in parallel
#     # TODO: passer en multiprocess plutôt que threaded ?
#     with ThreadPoolExecutor(max_workers=4) as _executor:
#         futures = [_executor.submit(_preprocess_and_ocr, img, o[0], o[1]) for o in operations]
#         return max([f.result() for f in concurrent.futures.as_completed(futures)], key=itemgetter(1))
#
#
# def _no_preprocess(img):
#     return img


# def _bilateral(img):
#     img = cv2.bilateralFilter(img, 5, 55, 60)
#     img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
#     _, img = cv2.threshold(img, 240, 255, 1)
#     return img
#
#
# def _adaptive(img):
#     imgf = cv2.adaptiveThreshold(img, 200, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 21, 2)
#     return imgf
#
#
# def _normalize(img):
#     # normalize image
#     norm_img = np.zeros((img.shape[0], img.shape[1]))
#     return cv2.normalize(img, norm_img, 0, 255, cv2.NORM_MINMAX)
#
#
# def _remove_noise(img):
#     img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
#     return cv2.fastNlMeansDenoisingColored(img, None, 10, 10, 7, 15)


def _image_to_blocks(img, lang="fra", config="--oem 3 --psm 1 " + tess_config):
    d = pytesseract.image_to_data(img, lang=lang, config=config, output_type='data.frame')
    # on filtre les données où la confiance est nulle ou négative
    d = d[d.conf > 0]
    # on filtre les données où le texte ne vaut rien
    if d.empty:
        return d
    try:
        d.text = d.text.str.strip()
    except Exception as e:
        raise e
    d.replace('', np.nan, inplace=True)
    d.dropna(inplace=True)

    if d.empty:
        return d
    # on regroupe par blocs
    lines = d.groupby(['block_num', 'par_num', 'line_num'])['text'].apply(list)
    # on calcule la confiance par bloc
    conf = d.groupby(['block_num', 'par_num', 'line_num'])['conf'].mean()
    # on calcule les coordonnées de bloc
    x1 = d.groupby(['block_num', 'par_num', 'line_num'])['left'].min()
    y1 = d.groupby(['block_num', 'par_num', 'line_num'])['top'].min()
    x2 = d.groupby(['block_num', 'par_num', 'line_num'])['left'].max() + \
         d.groupby(['block_num', 'par_num', 'line_num'])['width'].max()
    y2 = y1 + d.groupby(['block_num', 'par_num', 'line_num'])['height'].max()
    # on stocke le tout dans un nouveau dataframe
    output_df = pd.DataFrame({'text': lines, 'confidence': conf, 'x1': x1, 'y1': y1, 'x2': x2, 'y2': y2})
    return output_df


def _preprocess_and_ocr(img, preprocess, lang):
    img = preprocess(img)
    return _make_ocr(img, lang)


def _make_ocr(img, lang='eng'):
    final_text = ""
    final_confidence = 0

    df = _image_to_blocks(img, lang=lang)
    if df.empty:
        return "", 0
    for index, row in df.iterrows():
        actual_text = ' '.join(row['text']).strip('').replace('\n', '')
        if row['confidence'] < 90:
            new_image = img[row['y1']:row['y2'], row['x1']:row['x2']]
            new_row = _image_to_blocks(new_image, lang=lang, config="--oem 3 --psm 6" + tess_config)
            if new_row.empty:
                final_text += actual_text
                continue
            new_row = new_row.iloc[0]
            new_text = ' '.join(new_row['text']).strip('').replace('\n', '')
            new_confidence = new_row['confidence']
            if new_confidence > row['confidence']:
                final_text += " " + new_text + " "
            else:
                final_text += actual_text
            final_confidence += max(new_confidence, row['confidence'])
        else:
            final_text += actual_text
            final_confidence += row['confidence']
    final_confidence = final_confidence / df.shape[0]
    return final_text.strip(), final_confidence


from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import TypeAlias, Any, Union, Dict, List

import numpy
from fastapi_camelcase import CamelModel
from pydantic import BaseModel


class PropertyType(Enum):
    string = "string"
    number = "number"
    tag = "tag"
    multi_tags = "multi_tags"
    image_link = "image_link"
    url = "url"
    date = "date"
    path = "path"
    color = "color"
    checkbox = "checkbox"

    id = 'id'
    sha1 = 'sha1'
    ahash = 'ahash'
    folder = 'folder'
    width = 'width'
    height = 'height'


class PropertyMode(Enum):
    id = 'id'
    sha1 = 'sha1'


class Property(BaseModel):
    id: int
    name: str
    type: PropertyType
    mode: PropertyMode
    computed: bool = False


class PropertyDescription(Property):
    id: int | None = None
    type: PropertyType | None = None
    col: int


class PropertyUpdate(BaseModel):
    id: int
    name: str


# @dataclass(slots=True)
# class PropertyValue:
#     property_id: int
#
#     image_id: int
#     sha1: str
#
#     value: Any


@dataclass(slots=True)
class InstancePropertyValue:
    property_id: int
    instance_id: int
    value: Any


class ImagePropertyValue(CamelModel):
    property_id: int
    sha1: str
    value: Any


@dataclass(slots=True)
class Tag:
    id: int
    property_id: int
    value: str
    parents: list[int]
    color: int

    def __post_init__(self):
        self.value = str(self.value)


class TagUpdate(CamelModel):
    id: int
    value: str
    parent_id: list[int] | None = None
    color: int | None = None


@dataclass(slots=True)
class Instance:
    # Should be equal order to SQL
    id: int
    folder_id: int
    name: str
    extension: str

    sha1: str
    url: str
    height: int
    width: int
    ahash: str = ''

    properties: dict[int, InstancePropertyValue] = field(default_factory=dict)


# @dataclass(slots=True)
# class Image:
#     sha1: str
#     path: str
#     properties: dict[int, PropertyValue] = field(default_factory=dict)


@dataclass(slots=True)
class ImageImportTask:
    image_path: str
    folder_id: int


@dataclass(slots=True)
class ComputedValue:
    sha1: str
    ahash: str
    vector: numpy.ndarray


@dataclass(slots=True)
class Vector:
    source: str
    type: str
    sha1: str
    data: numpy.ndarray


class VectorDescription(CamelModel):
    source: str
    type: str
    count: int | None = None


class ProjectVectorDescriptions(CamelModel):
    vectors: list[VectorDescription] = []
    default_vectors: VectorDescription | None = None


class Parameters(BaseModel):
    folders: list[str]
    tabs: list[dict]


class Folder(BaseModel):
    id: int | None = None
    path: str
    name: str
    parent: int | None = None
    children: Dict[str, Folder] = {}


class Tab(BaseModel):
    id: int | None = None
    data: dict | None = None


class PathRequest(BaseModel):
    path: str


class UpdateCounter(CamelModel):
    action: int = 0
    image: int = 0


class StatusUpdate(CamelModel):
    tasks: List[TaskState] = []
    plugin_loaded: bool = False
    update: UpdateCounter = UpdateCounter()


class TaskState(BaseModel):
    name: str
    id: str
    total: int
    remain: int
    computing: int = 0
    done: bool = True


@dataclass(slots=True)
class Clusters:
    clusters: list[list[str]]
    distances: list[int]


class ActionContext(CamelModel):
    instance_ids: List[int] | None = None
    property_ids: List[int] | None = None
    file: str | None = None
    text: str | None = None
    ui_inputs: Dict[str, Any] = {}


class ParamDescription(CamelModel):
    name: str
    description: str | None = None
    type: str
    default_value: Any


class FunctionDescription(CamelModel):
    id: str
    name: str
    description: str | None = None
    action: str
    params: List[ParamDescription] = []


class PluginBaseParamsDescription(BaseModel):
    description: str | None = None
    params: List[ParamDescription] = []


class PluginDefaultParams(BaseModel):
    name: str
    # example: base.export_path = 'my/path'
    base: Dict[str, Any] = {}
    # example: functions.my_func.param1 = 'default_value'
    functions: Dict[str, Dict[str, Any]] = {}


class PluginDescription(CamelModel):
    name: str
    description: str | None = None
    path: str
    base_params: PluginBaseParamsDescription
    registered_functions: List[FunctionDescription] = []
    defaults: PluginDefaultParams


class ActionDescription(CamelModel):
    name: str
    selected_function: str | None = None
    available_functions: List[str] = []


class ActionParam(CamelModel):
    name: str
    value: str


class SetMode(Enum):
    set = 'set'
    add = 'add'
    delete = 'delete'


class ColumnOption(BaseModel):
    ignore:bool = False
    property_mode: PropertyMode | None = None


ImportOptions = dict[int, ColumnOption]



JSON: TypeAlias = Union[dict[str, "JSON"], list["JSON"], str, int, float, bool, None]
Tags: TypeAlias = dict[int, dict[int, Tag]]
# Properties: TypeAlias = dict[int, Property]



# TODO: conditionnal imports for people who just want base app
from __future__ import annotations

import os
import pickle
from hashlib import sha1 as sha1hash

import numpy as np
from PIL import Image
from imagehash import average_hash
from sklearn.decomposition import PCA
from sklearn.neighbors import KDTree

from .transformers import get_transformer
PCA_SIZE = 10
USE_PCA_IF_POSSIBLE = True

transformer = get_transformer("clip")

pca: None | PCA = None
tree: None | KDTree = None


def to_sha1(image: Image) -> str:
    return sha1hash(image.tobytes()).hexdigest()


def to_vector(image: Image, project_path: str):
    vec = transformer.to_vector(image)
    # if a pca was already trained use it
    if get_pca(project_path) is not None:
        return to_pca(vec)
    return vec


def to_average_hash(image: Image):
    ahash = average_hash(image)
    return ahash


# def to_ocr(image: Instance):
#     return full_ocr(image)

def get_pca(project_path: str):
    path = os.path.join(project_path, 'pca.pkl')
    global pca
    if pca:
        return pca
    if not pca and os.path.exists(path):
        with open(path, 'rb') as f:
            pca = pickle.load(f)
        return pca
    return None


def create_pca(vectors: [], project_path: str):
    from sklearn.decomposition import PCA
    global pca
    pca = PCA(PCA_SIZE)
    pca.fit(vectors)
    save_pca(project_path)


def to_pca(vector: np.ndarray):
    if pca is not None:
        return np.float32(pca.transform([vector])[0])


def save_pca(project_path: str):
    with open(os.path.join(project_path, 'pca.pkl'), 'wb') as f:
        pickle.dump(pca, f)


def can_compute_pca(nb_vectors: int, vector_sample: np.array) -> bool:
    """
    Can we start training a PCA to reduce vector size ?
    Basically it's: do we have more vectors than 110% the number of dimensions ?
    :param nb_vectors: number of vectors that are going to be used to train pca
    :param vector_sample: one of the vectors
    :return: bool
    """
    if not USE_PCA_IF_POSSIBLE:
        return False
    if get_pca() is not None:
        return False
    if nb_vectors > vector_sample.shape[0] + (vector_sample.shape[0] / 10):
        return True
    return False


import numpy as np
from PIL import Image
from transformers import MobileNetV2Model, AutoImageProcessor, CLIPModel, CLIPProcessor, CLIPTokenizer, logging

logging.set_verbosity_error()


class GoogleTransformer:
    def __init__(self):
        self.model = MobileNetV2Model.from_pretrained("google/mobilenet_v2_1.0_224")
        self.processor = AutoImageProcessor.from_pretrained("google/mobilenet_v2_1.0_224")

    @property
    def can_handle_text(self):
        return False

    def to_vector(self, image: Image) -> np.ndarray:
        input1 = self.processor(images=image, return_tensors="pt")
        output1 = self.model(**input1)
        pooled_output1 = output1[1].detach().numpy()
        vector = pooled_output1.flatten()
        return vector


class CLIPTransformer:
    def __init__(self):
        self.model = CLIPModel.from_pretrained("openai/clip-vit-base-patch32")
        self.processor = CLIPProcessor.from_pretrained("openai/clip-vit-base-patch32")
        self.tokenizer = CLIPTokenizer.from_pretrained("openai/clip-vit-base-patch32")

    @property
    def can_handle_text(self):
        return True

    def to_vector(self, image: Image) -> np.ndarray:
        image = self.processor(
            text=None,
            images=image,
            return_tensors="pt"
        )["pixel_values"]
        embedding = self.model.get_image_features(image)
        # convert the embeddings to numpy array
        embedding_as_np = embedding.cpu().detach().numpy()
        return embedding_as_np[0]

    def to_text_vector(self, text: str) -> np.ndarray:
        inputs = self.tokenizer(text, return_tensors="pt")
        text_embeddings = self.model.get_text_features(**inputs)
        # convert the embeddings to numpy array
        embedding_as_np = text_embeddings.cpu().detach().numpy()
        return embedding_as_np.reshape(1, -1)


def get_transformer(model="small"):
    if model == "small":
        return GoogleTransformer()
    elif model == "clip":
        return CLIPTransformer()
    else:
        return GoogleTransformer()


import logging

from PIL import Image

from panoptic.core.project.project import Project
from panoptic.core.task.task import Task
from panoptic.models import Instance, Vector

from . import compute
from .create_faiss_index import compute_faiss_index

logger = logging.getLogger('FaissPlugin:VectorTask')


class ComputeVectorTask(Task):
    def __init__(self, project: Project, source: str, type_: str, instance: Instance):
        super().__init__()
        self.project = project
        self.source = source
        self.type = type_
        self.instance = instance
        self.name = 'Clip Vectors'

    async def run(self):
        instance_id = self.instance.id
        instance = (await self.project.db.get_instances(ids=[instance_id]))[0]
        exist = await self.project.db.vector_exist(self.source, self.type, instance.sha1)
        if exist:
            return

        folder = await self.project.db.get_folder(instance.folder_id)
        file_path = f"{folder.path}/{instance.name}"
        vector_data = await self._async(self.compute_image, file_path, self.project.base_path)
        vector = Vector(self.source, self.type, instance.sha1, vector_data)
        res = await self.project.db.add_vector(vector)
        del vector
        # gc.collect()
        logging.info('computed image: ', instance_id, '  :  ', res.sha1)
        return res

    async def run_if_last(self):
        await compute_faiss_index(self.project.base_path, self.project.db, self.source, self.type)
        logging.info('computed faiss index')

    @staticmethod
    def compute_image(image_path: str, project_path: str):
        image = Image.open(image_path)
        image = image.convert('RGB')
        vector = compute.to_vector(image, project_path)

        del image
        # gc.collect()

        return vector



from __future__ import annotations

import csv
from collections import defaultdict

import pandas as pd
from random import randint
from typing import TYPE_CHECKING

from fastapi import UploadFile

if TYPE_CHECKING:
    from panoptic.core.project.project import Project
from panoptic.models import PropertyType, PropertyMode, PropertyDescription, Tag, Property, ImportOptions, SetMode


def parse_list(value: str):
    if value is None or value == '':
        return None
    # value = value.replace('[', '')
    # value = value.replace(']', '')
    return value.split(',')


def parse_header(index: int, name: str):
    name, remain = name.split('[')
    type_ = PropertyType(remain.split(']')[0])
    return index, name, type_


class Importer:
    def __init__(self, project: Project):
        self.project = project
        self._df: pd.DataFrame | None = None

    async def upload_csv(self, file: UploadFile):
        file_data = pd.read_csv(file.file, delimiter=';', encoding='utf-8')
        self._df = file_data

        return True

    async def import_file(self, options: ImportOptions):
        # TODO: refactor all this to use pandas methods properly
        # Read first row to determine properties
        columns = list(self._df.columns)
        file_key = columns[0]
        file_props = [parse_header(i + 1, col_name) for i, col_name in enumerate(columns[1:]) if col_name]
        file_props = [p for p in file_props if not (p[0] in options and options[p[0]].ignore)]

        # map of col index to property id
        col_to_prop: dict[int, Property] = {}

        # map col to existing property if possible
        db_props = await self.project.db.get_properties()
        for prop in db_props:
            for col_i, name, type_ in file_props:
                if name == prop.name and type_ == prop.type:
                    col_to_prop[col_i] = prop

        # create missing properties and add to map
        for col_i, name, type_ in file_props:
            if col_i not in col_to_prop:
                # use options info to determine property mode
                mode = options[col_i].property_mode if col_i in options else PropertyMode.id
                # fallback to id mode if None
                mode = PropertyMode.id if not mode else mode
                print('create property: ', name, type_, mode)
                new_prop = await self.project.db.add_property(name, type_, mode.value)
                col_to_prop[col_i] = new_prop

        # read full csv
        row_to_id: dict[int, int] = {}
        # map every row to and instance id
        if file_key == 'id':
            id_list = list(self._df.id)
            [row_to_id.update({i: id_}) for i, id_ in enumerate(id_list)]
        # if file path give map to an existing and empty instance or a new clone
        if file_key == '/':
            path_list = list(self._df.path)
            instances = await self.project.db.empty_or_clone(path_list)
            [row_to_id.update({i: instance.id}) for i, instance in enumerate(instances)]

        # for each property create arrays of (instance_id, value) pairs
        property_ids = [p.id for p in col_to_prop.values()]
        property_values: dict[int, list] = {i: [] for i in property_ids}

        for col_i in col_to_prop.keys():
            for r in self._df.iterrows():
                index, row = r
                property_values[col_to_prop[col_i].id].append((row_to_id[index], row[col_i]))

        # for each property set the property values
        for prop_id, pairs in property_values.items():
            ids, values = zip(*pairs)
            properties = list(col_to_prop.values())
            prop = properties[[p.id for p in properties].index(prop_id)]
            # if is tag or multi_tags property check tags first
            if prop.type == PropertyType.multi_tags or prop.type == PropertyType.tag:
                # map tag name to db_tag
                tags = await self.project.db.get_tags(prop_id)
                name_to_tag: dict[str, Tag] = {t.value: t for t in tags}
                values = [parse_list(v) for v in values]
                if prop.type == PropertyType.tag:
                    values = [[v[0]] if v else v for v in values]
                import_tags = set([t for v in values if v for t in v])

                # create missing tags
                to_create = [t for t in import_tags if t not in name_to_tag]
                for tag_name in to_create:
                    tag = await self.project.db.add_tag(prop_id, tag_name, 0, randint(0, 11))
                    name_to_tag[tag.value] = tag
                # replace tag names by tag ids
                values = [[name_to_tag[t].id for t in v] if v else None for v in values]
                for id_, value in zip(ids, values):
                    await self.project.db.set_tag_property_value(prop_id, [id_], value, SetMode.add)
            else:
                await self.project.db.set_property_values_array(prop_id, ids, values)

    async def analyse_file(self):
        # TODO: refactor all this to use pandas methods properly
        if self._df is None:
            raise Exception('No csv file was uploaded')

        # Read first row to determine properties
        columns = list(self._df.columns)
        file_key = columns[0]
        file_props = [parse_header(i + 1, col_name) for i, col_name in enumerate(columns[1:]) if col_name]
        properties = await self.project.db.get_properties(no_computed=True)
        col_to_prop: dict[int, PropertyDescription] = {}

        for i, name, type_ in file_props:
            col_to_prop[i] = PropertyDescription(name=name, type=type_, mode=PropertyMode.sha1, col=i)

        for prop in properties:
            for desc in col_to_prop.values():
                if desc.name == prop.name and desc.type == prop.type:
                    desc.id = prop.id
                    desc.mode = prop.mode

        import_props = list(col_to_prop.values())

        row_to_sha1: dict[int, str] = {}
        instances = await self.project.db.get_instances()
        key_desc = PropertyDescription(col=0, name='key', type=None, mode=PropertyMode.id)
        if file_key == 'id':
            key_desc.type = PropertyType.id
            key_desc.id = -1
            id_to_sha1 = {i.id: i.sha1 for i in instances}
            id_list = list(self._df.id)
            [row_to_sha1.update({i: id_to_sha1[id_]}) for i, id_ in enumerate(id_list)]
        if file_key == '/':
            key_desc.id = -7
            key_desc.type = PropertyType.path
            path_list = list(self._df.path)
            path_to_sha1 = {p: None for p in [path for path in path_list]}
            [path_to_sha1.update({i.url: i.sha1}) for i in instances if i.url in path_to_sha1]
            empties = [k for k in path_to_sha1.keys() if path_to_sha1[k] is None]
            if empties:
                raise Exception(f'{len(empties)} path(s) not found ' + ','.join([e for e in empties]))
            [row_to_sha1.update({i: path_to_sha1[path]}) for i, path in enumerate(path_list)]

        for prop in [p for p in import_props if p.id is None]:
            sha1_values = defaultdict(list)
            for r in self._df.iterrows():
                index, row = r
                values = sha1_values[row_to_sha1[index]]
                values.append(row.iloc[prop.col])
                if values[0] != row.iloc[prop.col]:
                    prop.mode = PropertyMode.id
                    break

        return [key_desc, *import_props]




import json
import os
from asyncio import sleep
from pathlib import Path
from typing import List

from pydantic import BaseModel

from panoptic.core.project.project import Project
from panoptic.utils import get_datadir


class ProjectId(BaseModel):
    name: str | None = None
    path: str | None = None


class PanopticData(BaseModel):
    projects: list[ProjectId]
    last_opened: ProjectId | None = None
    plugins: List[str] = []


class Panoptic:
    def __init__(self):
        self.global_file_path = get_datadir() / 'panoptic' / 'projects.json'
        self.data = self.load_data()
        self.project_id = None
        self.project: Project | None = None

        if not self.data.plugins:
            from panoptic.plugins import FaissPlugin
            module_path = os.path.abspath(FaissPlugin.__file__)
            module_path = module_path.replace('__init__.py', '')
            self.data.plugins.append(module_path)

    def load_data(self):
        try:
            with open(self.global_file_path, 'r') as file:
                data = json.load(file)
                return PanopticData(**data)
        except (FileNotFoundError, json.JSONDecodeError):
            return PanopticData(projects=[])

    def save_data(self):
        directory = os.path.dirname(self.global_file_path)
        # Create the directory if it doesn't exist
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(self.global_file_path, 'w') as file:
            json.dump(self.data.dict(), file, indent=2)

    async def create_project(self, name, path):
        if any(project.path == path for project in self.data.projects):
            raise f"A project_id with path '{path}' already exists."
        # else:
        if not os.path.exists(path):
            os.makedirs(path)
        project = ProjectId(name=name, path=path)
        self.data.projects.append(project)
        await self.load_project(path)

    async def import_project(self, path: str):
        p = Path(path)
        if not (p / 'panoptic.db').exists():
            raise ValueError('Folder is not a panoptic project_id (No panoptic.db file found)')
        if any(project.path == path for project in self.data.projects):
            raise f"ProjectId is already imported."
        project = ProjectId(path=str(p), name=str(p.name))
        self.data.projects.append(project)
        await self.load_project(path)

    def remove_project(self, path):
        self.data.projects = [p for p in self.data.projects if p.path != path]
        self.save_data()

    def rename_project(self, path, new_name):
        for project in self.data.projects:
            if project.path == path:
                project.name = new_name
        if self.data.last_opened and self.data.last_opened.path == path:
            self.data.last_opened.name = new_name
        self.save_data()

    async def load_project(self, path):
        for project in self.data.projects:
            if str(project.path) == str(path):
                self.save_data()
                self.project_id = project

                if self.project:
                    await self.project.close()
                self.project = Project(path, self.data.plugins)
                await self.project.start()

                from panoptic.routes.project_routes import set_project
                set_project(self.project)

    def add_plugin_path(self, path: str):
        if path in self.data.plugins:
            return
        init_path = Path(path) / '__init__.py'
        if not init_path.exists():
            raise Exception(f'No __init__.py file found at {path}')
        self.data.plugins.append(path)
        self.save_data()

    def del_plugin_path(self, path: str):
        if path in self.data.plugins:
            self.data.plugins.remove(path)
            self.save_data()

    def get_plugin_paths(self):
        return self.data.plugins

    async def close_project(self):
        self.project_id = None
        self.data.last_opened = {}
        self.save_data()
        await self.project.close()
        self.project = None
        from panoptic.routes.project_routes import set_project
        set_project(self.project)

    async def close(self):
        if self.project:
            await self.project.close()

    def is_loaded(self):
        return self.project_id is not None



import os.path

from pydantic import BaseModel

from panoptic.core.project.project import Project
from panoptic.models import Instance, ActionContext
from panoptic.models.results import GroupResult, Group, InstanceMatch, SearchResult
from panoptic.plugin import Plugin
from panoptic.utils import group_by_sha1
from .compute import reload_tree, get_similar_images, make_clusters
from .compute_vector_task import ComputeVectorTask


class FaissPluginParams(BaseModel):
    """
    Some base parameters for the Faiss Plugin
    @test_int: an int parameter
    @test_str: a str param
    """
    test_int: int = 0
    test_str: str = ''
    test_bool: bool = False


class FaissPlugin(Plugin):
    """
    Default Machine Learning plugin for Panoptic
    Uses CLIP to generate vectors and FAISS for clustering / similarity functions
    """

    def __init__(self, project: Project, plugin_path: str):
        super().__init__(name='Faiss', project=project, plugin_path=plugin_path)
        self.params = FaissPluginParams()
        reload_tree(project.base_path)

        project.on.import_instance.register(self.compute_image_vector)
        project.action.find_images.register(self, self.find_images)
        project.action.group_images.register(self, self.compute_clusters)

    async def start(self):
        await super().start()
        vectors = await self.project.db.get_vectors(self.name, 'clip')

        # TODO: handle this properly with an import hook
        if not os.path.exists(os.path.join(self.project.base_path, 'tree_faiss.pkl')) and len(vectors) > 0:
            from panoptic.plugins.FaissPlugin.create_faiss_index import compute_faiss_index
            await compute_faiss_index(self.project.base_path, self.project.db, self.name, 'clip')
            reload_tree(self.project.base_path)

    async def compute_image_vector(self, instance: Instance):
        task = ComputeVectorTask(self.project, self.name, 'clip', instance)
        self.project.task_queue.add_task(task)

    async def compute_clusters(self, context: ActionContext, nb_clusters: int = 10):
        """
        Computes images clusters with Faiss
        @nb_clusters: requested number of clusters
        """
        instances = await self.project.db.get_instances(context.instance_ids)
        sha1_to_instance = group_by_sha1(instances)
        sha1s = list(sha1_to_instance.keys())
        if not sha1s:
            return None

        vectors = await self.project.db.get_vectors(source=self.name, type_='clip', sha1s=sha1s)
        clusters, distances = make_clusters(vectors, method="kmeans", nb_clusters=nb_clusters)

        groups = [Group(ids=[i.id for sha1 in cluster for i in sha1_to_instance[sha1]], score=distance) for
                  cluster, distance in zip(clusters, distances)]
        return GroupResult(groups=groups)

    async def find_images(self, context: ActionContext):
        instances = await self.project.db.get_instances(context.instance_ids)
        sha1s = [i.sha1 for i in instances]
        ignore_sha1s = set(sha1s)
        vectors = await self.project.db.get_vectors(source=self.name, type_='clip', sha1s=sha1s)
        vector_datas = [x.data for x in vectors]
        res = get_similar_images(vector_datas)
        index = {r['sha1']: r['dist'] for r in res if r['sha1'] not in ignore_sha1s}

        res_sha1s = list(index.keys())
        res_instances = await self.project.db.get_instances(sha1s=res_sha1s)
        matches = [InstanceMatch(id=i.id, score=index[i.sha1]) for i in res_instances if i.sha1 in index]
        return SearchResult(matches=matches)



import sqlite3
import sys

def update_paths(db_path, new_root_path):
    # Connexion à la base de données SQLite
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()

    try:
        # Récupération du root_path à remplacer
        cursor.execute("SELECT path FROM folders WHERE id=1")
        old_root_path = cursor.fetchone()[0]
        
        print("old path " + old_root_path)
        # Calcul de la longueur du root_path pour découper les chemins
        old_root_len = len(old_root_path)

        # Mise à jour des chemins en remplaçant le root_path
        cursor.execute("UPDATE folders SET path = ? || substr(path, ?) WHERE path LIKE ?",
                       (new_root_path, old_root_len + 1, old_root_path + "%"))

        cursor.execute("UPDATE instances SET url = ? || substr(url, ?) WHERE url LIKE ?",
                       (new_root_path, old_root_len + 1, old_root_path + "%"))


        conn.commit()
        print("Les chemins ont été mis à jour avec succès.")

    except sqlite3.Error as e:
        print("Erreur lors de la mise à jour des chemins :", e)

    finally:
        # Fermeture de la connexion
        conn.close()

if __name__ == "__main__":
    # Vérification des arguments en ligne de commande
    if len(sys.argv) != 3:
        print("Usage: python script.py chemin_base_de_données nouveau_root_path")
        sys.exit(1)

    db_path = sys.argv[1]
    new_root_path = sys.argv[2]

    # Appel de la fonction pour mettre à jour les chemins
    update_paths(db_path, new_root_path)



import os
import webbrowser

import uvicorn
from fastapi import FastAPI
from starlette.middleware.cors import CORSMiddleware
from starlette.staticfiles import StaticFiles

from panoptic.core.panoptic import Panoptic
from panoptic.routes.project_routes import project_router
from panoptic.routes.project_selection_routes import selection_router, set_panoptic
from panoptic.utils import get_base_path


def start():
    panoptic = Panoptic()
    panoptic.load_data()

    HOST = os.getenv("PANOPTIC_HOST", None)
    # default port for Panoptic backend
    PORT = os.getenv("PANOPTIC_PORT", 8000)

    # FastAPI setup
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    BASE_PATH = get_base_path()
    # base path for the static folder

    app.include_router(selection_router)
    app.include_router(project_router)

    set_panoptic(panoptic)

    # static directory route
    app.mount("/", StaticFiles(directory=os.path.join(BASE_PATH, "html"), html=True), name="static")

    dev_url = 'http://localhost:5173/'
    prod_url = 'http://localhost:' + str(PORT) + '/'
    front_url = dev_url if os.getenv("PANOPTIC_ENV", "PROD") == "DEV" else prod_url

    if not os.environ.get('REMOTE'):
        webbrowser.open(front_url)

    @app.on_event("shutdown")
    async def shutdown_event():
        await panoptic.close()

    uvicorn.run(app, host=HOST, port=PORT)


if __name__ == '__main__':
    start()



from pathlib import Path

from pydantic import BaseModel

from panoptic.core.project.project import Project
from panoptic.models import ActionContext
from panoptic.plugin import Plugin


class TestParams(BaseModel):
    eau: str = None
    terre: int = 0
    feu: float = 2.4
    air: Path = None


class PluginBase(Plugin):
    def __init__(self, project: Project, plugin_path: str):
        super().__init__(name='TestPlugin1', project=project, plugin_path=plugin_path)
        self.params = TestParams()

        self.project.action.group_images.register(self, self.firefly)

    async def firefly(self, context: ActionContext, damage: int):
        pass



import json
from collections import defaultdict
from random import randint
from typing import Any, Dict

from panoptic.core.db.db import Db
from panoptic.core.db.db_connection import DbConnection
from panoptic.core.project.project_events import ImportInstanceEvent
from panoptic.models import Property, PropertyUpdate, PropertyType, InstancePropertyValue, Instance, Tag, \
    TagUpdate, Vector, PluginDefaultParams, VectorDescription, ActionParam, ProjectVectorDescriptions, SetMode, \
    PropertyMode
from panoptic.models.computed_properties import computed_properties
from panoptic.models.results import DeleteTagResult
from panoptic.utils import convert_to_instance_values, clean_value, get_computed_values


class ProjectDb:
    def __init__(self, conn: DbConnection):
        self._db = Db(conn)

        self.on_import_instance = ImportInstanceEvent()

    async def close(self):
        await self._db.close()

    def get_raw_db(self):
        return self._db

    # =====================================================
    # =================== Properties ======================
    # =====================================================

    async def add_property(self, name: str, property_type: PropertyType, mode='id') -> Property:
        return await self._db.add_property(name, property_type.value, mode)

    async def get_properties(self, no_computed=False) -> list[Property]:
        properties = await self._db.get_properties()
        if no_computed:
            return properties
        return [*properties, *computed_properties.values()]

    async def update_property(self, update: PropertyUpdate) -> Property:
        existing_property = await self._db.get_property(update.id)
        if not existing_property:
            raise Exception(f'Property {update.id} does not exist')
        new_property = existing_property.copy(update=update.dict(exclude_unset=True))
        await self._db.update_property(new_property)
        return new_property

    async def delete_property(self, property_id: str):
        await self._db.delete_property(property_id)

    # =====================================================
    # =============== Property Values =====================
    # =====================================================

    async def get_property_values(self, instances: list[Instance], property_ids: list[int] = None, no_computed=False) \
            -> list[InstancePropertyValue]:
        instance_ids = [i.id for i in instances]
        sha1s = list({img.sha1 for img in instances})
        instance_values = await self._db.get_instance_property_values(property_ids=property_ids,
                                                                      instance_ids=instance_ids)
        image_values = await self._db.get_image_property_values(property_ids=property_ids, sha1s=sha1s)
        converted_values = convert_to_instance_values(image_values, instances)

        computed_ids = [] if no_computed else computed_properties.keys()
        if computed_ids and property_ids:
            computed_ids = [pId for pId in property_ids if pId < 0]
        computed_values = [v for i in instances for v in get_computed_values(i, computed_ids)]
        return [*instance_values, *converted_values, *computed_values]

    async def set_property_values(self, property_id: int, instance_ids: list[int], value: Any):
        prop = await self._db.get_property(property_id)
        value = clean_value(prop, value)
        # print(property_id, value)
        if prop.mode == PropertyMode.id:
            # print('mode id')
            if value is None:
                await self._db.delete_instance_property_value(property_id, instance_ids)
            else:
                await self._db.set_instance_property_value(property_id, instance_ids, value)
        if prop.mode == PropertyMode.sha1:
            # print('mode sha1')
            instances = await self._db.get_instances(ids=instance_ids)
            sha1s = {i.sha1 for i in instances}
            if value is None:
                await self._db.delete_image_property_value(property_id, list(sha1s))
            else:
                await self._db.set_image_property_value(property_id, list(sha1s), value)

        return [InstancePropertyValue(property_id=property_id, instance_id=i, value=value) for i in instance_ids]

    async def set_property_values_array(self, property_id: int, instance_ids: list[int], values: list[Any]):
        prop = await self._db.get_property(property_id)
        values = [clean_value(prop, v) for v in values]
        pairs = [(i, v) for i, v in zip(instance_ids, values)]
        to_delete = [p for p in pairs if p[1] is None]
        to_update = [p for p in pairs if p[1] is not None]
        if prop.mode == PropertyMode.id:
            if to_update:
                ids = [v[0] for v in to_update]
                vals = [v[1] for v in to_update]
                await self._db.set_instance_property_array_values(property_id=prop.id, instance_ids=ids, values=vals)
            if to_delete:
                ids = [v[0] for v in to_delete]
                await self._db.delete_instance_property_value(property_id=prop.id, instance_ids=ids)
        if prop.mode == PropertyMode.sha1:
            if to_update:
                ids = [v[0] for v in to_update]
                instances = await self._db.get_instances(ids=ids)
                index = {i.id: i.sha1 for i in instances}
                sha1_pairs = [(index[v[0]], v[1]) for v in to_update]
                sha1s = [p[0] for p in sha1_pairs]
                vals = [p[1] for p in sha1_pairs]
                await self._db.set_image_property_array_values(property_id=prop.id, sha1s=sha1s, values=vals)
            if to_delete:
                ids = [v[0] for v in to_delete]
                instances = await self._db.get_instances(ids=ids)
                sha1s = list({i.sha1 for i in instances})
                await self._db.delete_image_property_value(property_id=prop.id, sha1s=sha1s)
        return [InstancePropertyValue(property_id=property_id, instance_id=p[0], value=p[1]) for p in pairs]

    async def set_tag_property_value(self, property_id: int, instance_ids: list[int], value: list[int], mode: SetMode):
        if mode == SetMode.set:
            return await self.set_property_values(property_id=property_id, instance_ids=instance_ids, value=value)

        instances = await self.get_instances_with_properties(instance_ids=instance_ids, property_ids=[property_id])

        to_set = [i for i in instances if property_id not in i.properties]
        if to_set:
            set_res = await self.set_property_values(property_id=property_id, instance_ids=[i.id for i in to_set], value=value)
        else:
            set_res = []
        if mode == SetMode.add:
            to_add = [i for i in instances if property_id in i.properties]
            ids = [i.id for i in to_add]
            vals = [[*i.properties[property_id].value, *value] for i in to_add]
            vals = [[*{*v}] for v in vals]
            res_add = await self.set_property_values_array(property_id=property_id, instance_ids=ids, values=vals)
            return [*set_res, *res_add]

        if mode == SetMode.delete:
            to_del = [i for i in instances if property_id in i.properties]
            ids = [i.id for i in to_del]
            vals = [[v for v in i.properties[property_id].value if v not in value] for i in to_del]
            res_del = await self.set_property_values_array(property_id=property_id, instance_ids=ids, values=vals)
            return [*set_res, *res_del]

    # =====================================================
    # =================== Instances =======================
    # =====================================================

    async def add_instance(self, folder_id: int, name: str, extension: str, sha1: str, url: str, width: int,
                           height: int, ahash: str):
        res = await self._db.add_instance(folder_id, name, extension, sha1, url, width, height, ahash)
        self.on_import_instance.emit(res)

    async def get_instances(self, ids: list[int] = None, sha1s: list[str] = None):
        images = await self._db.get_instances(ids=ids, sha1s=sha1s)
        return images

    async def get_instances_with_properties(self, instance_ids: list[int] = None, property_ids: list[int] = None) \
            -> list[Instance]:
        instances = await self._db.get_instances(instance_ids)
        instance_values = await self.get_property_values(instances=instances, property_ids=property_ids)
        instance_index: dict[int, Instance] = {i.id: i for i in instances}
        [instance_index[v.instance_id].properties.update({v.property_id: v}) for v in instance_values]

        return instances

    async def empty_or_clone(self, paths: list[str]) -> list[Instance]:
        instances = await self.get_instances()
        path_to_instances = defaultdict(list)
        res: list[Instance] = []
        for i in instances:
            path_to_instances[i.url].append(i)
        for path in paths:
            if not path_to_instances[path]:
                raise Exception('{path} is not already imported in panoptic. '
                                'Impossible to import the data before importing the file')
            matches = sorted(path_to_instances[path], key=lambda x: x.id)
            values = await self.get_property_values(instances=[matches[0]], no_computed=True)
            if not values:
                res.append(matches[0])
            else:
                new_instance = await self._db.clone_instance(matches[0])
                res.append(new_instance)
        return res

    # ========== Tags ==========
    async def get_tags(self, prop: int = None) -> list[Tag]:
        tag_list = await self._db.get_tags(prop)
        return tag_list

    async def add_tag(self, property_id, value, parent_id, color: int) -> Tag:
        existing_tag = await self._db.get_tag(property_id, value)
        if existing_tag is not None:
            if await self._db.tag_in_ancestors(existing_tag.id, parent_id):
                raise Exception("Adding a tag that is an ancestor of himself")
            existing_tag.parents = list({*existing_tag.parents, parent_id})
            await self._db.update_tag(existing_tag)
            tag_id = existing_tag.id
        else:
            parents = [parent_id]
            tag_id = await self._db.add_tag(property_id, value, json.dumps(parents), color)
        return await self._db.get_tag_by_id(tag_id)

    async def add_tag_parent(self, tag_id, parent_id):
        tag = await self._db.get_tag_by_id(tag_id)
        parent = await self._db.get_tag_by_id(parent_id)
        if tag is None:
            raise Exception(f"Tag: {tag_id} doesnt exist")
        if parent is None:
            raise Exception(f"Parent Tag: {parent_id} doesnt exist")
        if await self._db.tag_in_ancestors(tag.id, parent_id):
            raise Exception("Adding a tag that is an ancestor of himself")
        tag.parents = list({*tag.parents, parent_id})
        tag.color = parent.color
        await self._db.update_tag(tag)
        return tag

    async def update_tag(self, update: TagUpdate) -> Tag:
        existing_tag = await self._db.get_tag_by_id(update.id)
        if not existing_tag:
            raise Exception("Trying to modify non existent tag")
        if update.color is not None:
            existing_tag.color = update.color
        if update.value is not None:
            existing_tag.value = update.value
        await self._db.update_tag(existing_tag)
        return existing_tag

    async def delete_tag(self, tag_id: int):
        tag = await self._db.get_tag_by_id(tag_id)
        instances = await self._db.get_instances()
        values = await self.get_property_values(instances, [tag.property_id])

        to_update = [v for v in values if tag_id in v.value]
        [v.value.remove(tag_id) for v in to_update if tag_id in v.value]

        ids = [v.instance_id for v in to_update]
        vals = [v.value for v in to_update]
        await self.set_property_values_array(tag.property_id, ids, vals)

        tags = await self._db.get_tags(tag.property_id)
        updated_tags = [t for t in tags if tag_id in t.parents]
        [t.parents.remove(tag_id) for t in updated_tags]
        [await self._db.update_tag(t) for t in updated_tags]

        await self._db.delete_tag_by_id(tag_id)
        # print(to_update, updated_tags)
        return DeleteTagResult(tag_id=tag.property_id, updated_values=to_update, updated_tags=updated_tags)

    async def delete_tag_parent(self, tag_id: int, parent_id: int) -> list[int]:
        tag = await self._db.get_tag_by_id(tag_id)
        tag.parents.remove(parent_id)
        color = randint(0, 11)
        tag.color = color
        await self._db.update_tag(tag)
        return tag

    # =========== Folders ===========
    async def add_folder(self, path: str, name: str, parent: int = None):
        return await self._db.add_folder(path, name, parent)

    async def get_folders(self):
        return await self._db.get_folders()

    async def get_folder(self, folder_id: int):
        return await self._db.get_folder(folder_id)

    # =========== Vectors ===========
    async def get_vectors(self, source: str, type_: str, sha1s: list[str] = None):
        return await self._db.get_vectors(source, type_, sha1s)

    async def get_default_vectors(self, sha1s: list[str] = None):
        default = await self._db.get_action_param('get_vectors')
        if not default:
            raise Exception('No default vectors set. Make sure vectors are computed')
        default = VectorDescription(**json.loads(default.value))
        return await self._db.get_vectors(default.source, default.type, sha1s)

    async def add_vector(self, vector: Vector):
        return await self._db.add_vector(vector)

    async def vector_exist(self, source: str, type_: str, sha1: str) -> bool:
        return await self._db.vector_exist(source, type_, sha1)

    async def set_default_vectors(self, vector: VectorDescription):
        await self._db.set_action_param(ActionParam(name='get_vectors', value=json.dumps(vector.json())))

    async def get_vectors_info(self):
        vectors = await self._db.get_vector_descriptions()
        default = await self._db.get_action_param('get_vectors')
        if default:
            default = VectorDescription(**json.loads(default.value))
        return ProjectVectorDescriptions(vectors=vectors, default_vectors=default)

    # ============ Plugins =============
    async def set_plugin_default_params(self, params: PluginDefaultParams):
        return await self._db.set_plugin_default_params(params)



import asyncio
import logging
import sys
import traceback
from concurrent.futures import Executor
from typing import Dict, List

from panoptic.core.task.task import Task
from panoptic.models import TaskState

logger = logging.getLogger('TaskQueue')


class TaskQueue:
    def __init__(self, executor: Executor, num_workers: int = 1):
        self._executor = executor
        self._priority_tasks = asyncio.Queue()
        self._tasks = asyncio.Queue()
        self._num_workers = num_workers
        self._workers: List[asyncio.Task] = []
        self._working: Dict[int, bool] = {}

        self._task_states: Dict[str, TaskState] = {}

        for i in range(num_workers):
            worker = asyncio.create_task(self._worker(i))
            self._workers.append(worker)

    def get_task_states(self) -> List[TaskState]:
        return list(self._task_states.values())

    def empty(self):
        return self._priority_tasks.empty() and self._tasks.empty()

    def add_task(self, task: Task):
        if task.has_priority:
            self._priority_tasks.put_nowait(task)
        else:
            self._tasks.put_nowait(task)

        if task.get_id() not in self._task_states:
            self._task_states[task.get_id()] = TaskState(id=task.get_id(), name=task.name, total=0, remain=0)

        state = self._task_states[task.get_id()]
        state.done = False
        state.total += 1
        state.remain += 1

    async def _worker(self, worker_id: int):
        while True:
            try:
                # if nothing to do mark as inactive
                if self.empty():
                    self._working[worker_id] = False

                # if we have priority tasks take it
                if not self._priority_tasks.empty():
                    task = await self._priority_tasks.get()
                    state = self._task_states[task.get_id()]
                    state.remain -= 1
                # else normal task
                else:
                    if self._tasks.empty():
                        await asyncio.sleep(0.5)  # sleep a bit if no tasks available
                        continue
                    task = await self._tasks.get()
                    state = self._task_states[task.get_id()]
                    state.remain -= 1

                # set working Flag and execute task
                self._working[worker_id] = True
                task.set_executor(self._executor)
                try:
                    state.computing += 1
                    await task.run()
                    state.computing -= 1
                except Exception as e:
                    state.computing -= 1
                    exc_type, exc_value, exc_traceback = sys.exc_info()
                    logger.error("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
                    logger.error(e)
                if state.remain == 0 and state.computing == 0:
                    state.done = True
                    await task.run_if_last()

            except Exception as e:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                logger.error("".join(traceback.format_exception(exc_type, exc_value, exc_traceback)))
                logger.error(e)

    async def close(self):
        for worker in self._workers:
            worker.cancel()
        await asyncio.gather(*self._workers, return_exceptions=True)



import inspect
import logging
from pathlib import Path
from typing import Any, List, Dict

from panoptic.models import ActionContext, FunctionDescription, ParamDescription, ActionDescription, ActionParam
from panoptic.plugin import Plugin
from panoptic.utils import AsyncCallable, to_str_type


def get_params(f):
    signature = inspect.signature(f)
    parameters = signature.parameters

    res = {}
    for p in parameters:
        param = parameters[p]
        type_ = param.annotation
        default = param.default
        if default is inspect.Parameter.empty:
            default = None
        res[p] = (type_, default)
    return res


possible_inputs = [int, float, str, List[str], bool, Path]


def get_param_description(f: AsyncCallable, param_name: str):
    doc: str = f.__doc__
    token = f'@{param_name}: '
    if not doc or token not in doc:
        return None

    token_start = doc.index(token)
    start = token_start + len(token)
    if '@' in doc[start:]:
        end = doc.index('@', start)
        return doc[start:end]
    return doc[start:]


def get_params_description(f: AsyncCallable) -> List[ParamDescription]:
    types = get_params(f)
    return [ParamDescription(name=t, type=to_str_type(types[t][0]), default_value=types[t][1], description=get_param_description(f, t))
            for t in types if types[t][0] in possible_inputs]


def get_registered_function_description(function_id: str, function: AsyncCallable, action_name: str):
    name = function.__name__
    description = function.__doc__
    if description and '@' in description:
        description = description[0: description.index('@')]
    params = get_params_description(function)
    return FunctionDescription(id=function_id, name=name, description=description, action=action_name, params=params)


# def verify_dependencies(f: AsyncCallable, required: List):
#     dependencies = get_dependencies(f).values()
#
#     if Instances in dependencies and Images in dependencies:
#         return False
#
#     count = 0
#     required_nb = len(required)
#     if Instances in required and Images in required:
#         required_nb -= 1
#
#     for r in required:
#         if r in dependencies:
#             count += 1
#
#     return count == required_nb


class Action:
    def __init__(self, id_: str, source: Any, function: AsyncCallable):
        self.id = id_
        self.source = source
        self._function = function

    async def call(self, *args, **kwargs):
        return await self._function(*args, **kwargs)


class ProjectAction:
    def __init__(self, name: str):
        self._possible_actions: List[Action] = []
        self._selected_action = None
        self.name = name

    def can_call(self):
        return self._selected_action is not None

    def register(self, source: Plugin, function: AsyncCallable):
        id_ = source.name + '.' + function.__name__
        action = Action(source=source, function=function, id_=id_)
        self._possible_actions.append(action)

        function_id = f'{source.name}.{function.__name__}'
        function_description = get_registered_function_description(function_id, function, self.name)
        source.registered_functions.append(function_description)
        if not self.can_call():
            self._selected_action = 0

    async def call(self, context: ActionContext, function: str = None):
        if function:
            action = [a for a in self._possible_actions if a.id == function][0]
        else:
            action = self._possible_actions[self._selected_action]
        return await action.call(context, **context.ui_inputs)

    def select_function(self, function_id):
        valid = [f.id for f in self._possible_actions]
        if function_id not in valid:
            logging.warning(f'{function_id} is not available in action {self.name}')
            return
        self._selected_action = valid.index(function_id)

    def get_description(self):
        selected = self._possible_actions[self._selected_action].id if self.can_call() else None
        available = [a.id for a in self._possible_actions]
        return ActionDescription(name=self.name, selected_function=selected, available_functions=available)


class ProjectActions:
    def __init__(self):
        self.filter_images = ProjectAction('filter')
        self.group_images = ProjectAction('group')
        self.find_images = ProjectAction('find_similar')
        self.action_images = ProjectAction('action_images')
        self.action_group = ProjectAction('action_group')
        self.import_properties = ProjectAction('import')
        self.export_properties = ProjectAction('export')

        action_list = [self.find_images, self.group_images, self.filter_images, self.action_images, self.action_group,
                       self.import_properties, self.export_properties]
        self.actions: Dict[str, ProjectAction] = {a.name: a for a in action_list}

    def get_actions_description(self):
        return [a.get_description() for a in self.actions.values()]

    def _set_action_function(self, action_name: str, function_id: str):
        action = self.actions[action_name]
        action.select_function(function_id)

    def set_action_functions(self, actions: List[ActionParam]):
        [self._set_action_function(action.name, action.value) for action in actions if action.name in self.actions]



import os
import pickle


def load_data(data_path):
    full_path = os.path.join(data_path, 'data.pkl')
    if full_path and os.path.exists(full_path):
        with open(full_path, 'rb') as f:
            data = pickle.load(f)
        return data
    return {
        'images': {},
        'pca': None,
        'tree': None
    }


def save_data(data_path: str, images_dict):
    full_path = os.path.join(data_path, 'data.pkl')
    with open(full_path, 'wb') as f:
        pickle.dump(images_dict, f)


def load_similarity_tree(path: str):
    path = os.path.join(path, 'tree_faiss.pkl')
    if not os.path.exists(path):
        return None
    with open(path, 'rb') as f:
        return pickle.load(f)



from __future__ import annotations

import datetime
import os
import shutil
from typing import List, TYPE_CHECKING

import pandas as pd
if TYPE_CHECKING:
    from panoptic.core.project.project import Project
from panoptic.models import PropertyType, Instance, Property


def _copy_images(images: List[Instance], destination_folder: str) -> None:
    os.makedirs(destination_folder, exist_ok=True)
    for image in images:
        url = image.url.replace("/images/", "", 1)
        try:
            shutil.copy(url, destination_folder)
        except BaseException:
            print(f'File {url} not found !')


def get_name(p: Property):
    if p.id == -1:
        return p.name
    return f'{p.name}[{p.type.value}]'


class Exporter:
    def __init__(self, project: Project):
        self.project = project

    async def export_data(self, path, name: str = None, instance_ids: [int] = None, properties=None,
                          copy_images: bool = False) -> str:
        if not name:
            name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        base_export_folder = os.path.join(path, 'exports')
        # Create the export folder if it doesn't exist
        os.makedirs(base_export_folder, exist_ok=True)

        export_folder = os.path.join(base_export_folder, name)
        if os.path.exists(export_folder):
            shutil.rmtree(export_folder)
        os.makedirs(export_folder)

        instances = await self.project.db.get_instances_with_properties(instance_ids)
        # Create a CSV file with the data
        if properties:
            data_file_path = os.path.join(export_folder, 'data.csv')
            df = await self._build_export_data(instances, properties)
            df.to_csv(data_file_path, index=False)

        if copy_images:
            # Create a folder for images and copy them
            image_folder_path = os.path.join(export_folder, 'images')
            _copy_images(instances, image_folder_path)

        return export_folder

    async def _build_export_data(self, images: [Instance], properties_list: list[int]):
        """
        Allow to export selected images and properties into a csv file
        """
        properties = await self.project.db.get_properties()
        tags = await self.project.db.get_tags()
        tag_index = {t.id: t for t in tags}

        folders = await self.project.db.get_folders()
        folder_index = {f.id: f for f in folders}

        # filter properties id that we want to keep
        id_to_prop = {p.id: p for p in properties}
        columns = [get_name(id_to_prop[p]) for p in properties_list]
        rows = []
        for image in images:
            row = []
            for prop_id in properties_list:
                prop = id_to_prop[prop_id]
                if prop.id in image.properties:
                    value = image.properties[prop.id].value
                    # if it's a tag let's fetch tag value from tag id
                    if prop.type == PropertyType.tag or prop.type == PropertyType.multi_tags:
                        # print('is tag !!')
                        if type(value) != list:
                            row.append(None)
                            continue
                        row.append(",".join([tag_index[t].value for t in value]))
                    elif prop.type == PropertyType.folder:
                        row.append(folder_index[value].name)
                    else:
                        row.append(value)
                else:
                    row.append(None)
            rows.append(row)
        df = pd.DataFrame.from_records(rows, columns=columns)
        return df



import asyncio
import os
import pathlib
import sys
from typing import List, Any, Callable, Awaitable, Dict

from pydantic import BaseModel

from panoptic.models import ParamDescription, Instance, ImagePropertyValue, InstancePropertyValue, Property
from panoptic.models.computed_properties import ComputedId


def get_datadir() -> pathlib.Path:
    """
    Returns a parent directory path
    where persistent application data can be stored.

    # linux: ~/.local/share
    # macOS: ~/Library/Application Support
    # windows: C:/Users/<USER>/AppData/Roaming
    # gently taken from https://stackoverflow.com/questions/19078969/python-getting-appdata-folder-in-a-cross-platform-way
    """

    home = pathlib.Path.home()
    if os.getenv('IS_DOCKER', False):
        return pathlib.Path('/data')
    elif sys.platform == "win32":
        return home / "AppData/Roaming"
    elif sys.platform == "linux":
        return home / ".local/share"
    elif sys.platform == "darwin":
        return home / "Library/Application Support"


def get_base_path():
    if getattr(sys, 'frozen', False):
        # Le programme est exécuté en mode fichier unique
        return sys._MEIPASS
    else:
        # Le programme est exécuté en mode script
        return os.path.dirname(__file__)


AsyncCallable = Callable[[Any, Any, Any], Awaitable[None]]


class EventListener:
    def __init__(self):
        self.callbacks: List[AsyncCallable] = []
        self._tasks = []

    def register(self, callback: AsyncCallable):
        # print('register')
        self.callbacks.append(callback)
        # print(self.callbacks)

    def redirect(self, listener):
        async def redirect(x):
            listener.emit(x)

        self.register(redirect)

    def unregister(self, callback: AsyncCallable):
        if callback in self.callbacks:
            self.callbacks.remove(callback)

    def emit(self, event: Any):
        # print(f'callbacks: {self.callbacks}')
        async def async_emit():
            # print(f'async callbacks: {self.callbacks}')
            # print(f'execute {len(self.callbacks)} callbacks')
            for callback in self.callbacks:
                await callback(event)

        task = asyncio.create_task(async_emit())
        self._tasks.append(task)
        task.add_done_callback(self._tasks.remove)


def to_str_type(type_):
    if type_ is int:
        return 'int'
    if type_ is float:
        return 'float'
    if type_ is str:
        return 'str'
    if type_ is List[str]:
        return '[str]'
    if type_ is bool:
        return 'bool'
    if type_ is pathlib.Path:
        return 'path'
    return None


def get_param_comment_from_model(model: BaseModel, param_name: str):
    doc: str = model.__doc__
    token = f'@{param_name}: '
    if not doc or token not in doc:
        return None

    token_start = doc.index(token)
    start = token_start + len(token)
    if '@' in doc[start:]:
        end = doc.index('@', start)
        return doc[start:end]
    return doc[start:]


def get_model_params_description(model: BaseModel):
    res: List[ParamDescription] = []
    fields = model.__fields__
    for field_name, field_info in fields.items():
        field_type = field_info.annotation
        default_value = field_info.default
        description = get_param_comment_from_model(model, field_name)
        res.append(ParamDescription(name=field_name, description=description, type=to_str_type(field_type),
                                    default_value=default_value))
    return res


def group_by_sha1(instances: List[Instance]):
    res: Dict[str, List[Instance]] = {}
    for i in instances:
        if i.sha1 not in res:
            res[i.sha1] = []
        res[i.sha1].append(i)
    return res


def convert_to_instance_values(values: list[ImagePropertyValue], instances: list[Instance]) \
        -> list[InstancePropertyValue]:
    sha1_to_instances = {i.sha1: [] for i in instances}
    [sha1_to_instances[i.sha1].append(i) for i in instances]

    def converter2(value: ImagePropertyValue):
        targets = sha1_to_instances[value.sha1]
        return [InstancePropertyValue(instance_id=i.id, property_id=value.property_id, value=value.value) for i in targets]

    pre_res = [converter2(v) for v in values]
    return [v for vv in pre_res for v in vv]


def clean_value(prop: Property, v: Any):
    # avoid setting to None
    if v == 0:
        return 0

    if not v:
        return None

    return v


def get_computed_values(instance: Instance, computed_ids: list[int]):
    res = []
    for i in computed_ids:
        if i == ComputedId.id:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.id))
        if i == ComputedId.sha1:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.sha1))
        if i == ComputedId.ahash:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.ahash))
        if i == ComputedId.folder:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.folder_id))
        if i == ComputedId.width:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.width))
        if i == ComputedId.height:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.height))
        if i == ComputedId.path:
            res.append(InstancePropertyValue(instance_id=instance.id, property_id=i, value=instance.url))

    return res


def all_equal(l: list):
    if len(l) == 1:
        return True
    for i in range(len(l) - 1):
        if l[i] != l[i + 1]:
            return False
    return True



from panoptic.core.db.db import Db
from panoptic.models import UpdateCounter


class ProjectUi:
    def __init__(self, db: Db):
        self._db = db
        self.update_counter = UpdateCounter()

    async def get_tabs(self):
        return await self._db.get_tabs()

    async def add_tab(self, data):
        return await self._db.add_tab(data)

    async def update_tab(self, data):
        return await self._db.update_tab(data)

    async def delete_tab(self, tab_id: int):
        return await self._db.delete_tab(tab_id)



from __future__ import annotations

from typing import List, Any, Dict
from typing import TYPE_CHECKING

from panoptic.models import PluginBaseParamsDescription, FunctionDescription, PluginDescription, PluginDefaultParams
from panoptic.utils import get_model_params_description

if TYPE_CHECKING:
    from panoptic.core.project.project import Project


class Plugin:
    def __init__(self, name: str, project: Project, plugin_path: str):
        self.params: Any | None = None
        self.name: str = name
        self.project = project
        self.registered_functions: List[FunctionDescription] = []
        self.path = plugin_path

    async def start(self):
        db_defaults = await self.project.db.get_raw_db().get_plugin_default_params(self.name)
        self.update_default_values(db_defaults)

    def update_default_values(self, params: PluginDefaultParams):
        if not self.params:
            return

        for param in params.base:
            if params.base[param]:
                setattr(self.params, param, params.base[param])

    def _get_param_description(self, db_base: Dict[str, Any]):
        if not self.params:
            return PluginBaseParamsDescription()
        description = self.params.__doc__
        if description and '@' in description:
            description = description[0: description.index('@')]

        params = get_model_params_description(self.params)
        # print('params', params)
        for p in params:
            if p.name in db_base and db_base[p.name]:
                p.default_value = db_base[p.name]
        return PluginBaseParamsDescription(description=description, params=params)

    async def get_description(self):
        name = self.name
        description = self.__doc__
        path = self.path

        defaults = PluginDefaultParams(name=name)
        db_defaults = await self.project.db.get_raw_db().get_plugin_default_params(plugin_name=name)

        base_params = self._get_param_description(db_defaults.base)
        # print(base_params)

        for param in base_params.params:
            defaults.base[param.name] = param.default_value

        for function in self.registered_functions:
            defaults.functions[function.name] = {}
            for param in function.params:
                defaults.functions[function.name][param.name] = param.default_value
                if function.name in db_defaults.functions and param.name in db_defaults.functions[function.name]:
                    defaults.functions[function.name][param.name] = db_defaults.functions[function.name][param.name]

        res = PluginDescription(name=name, description=description, path=path, base_params=base_params,
                                registered_functions=self.registered_functions, defaults=defaults)
        return res



import logging

from panoptic.core.project.project_db import ProjectDb
from .compute.similarity import create_similarity_tree_faiss

logger = logging.getLogger('Create Faiss')


async def compute_faiss_index(path: str, db: ProjectDb, source: str, type_: str):
    logger.info('Start')
    vectors = await db.get_vectors(source, type_)
    create_similarity_tree_faiss(path, vectors)
    logger.info('Success')

